<?php
/**
 * DB Class
 *
 * @since 1.0.0
 */

namespace WPO\WC\Product_Batch_Numbers\Database;

use WC_DateTime;
use DateTimeZone;
use Exception;

/**
 * If this file is called directly, abort.
 *
 * @since 1.0.0
 */
if ( ! defined( 'ABSPATH' ) ) {
	die;
}

if ( ! class_exists( 'wcpbnAbstractDb' ) ) {

	class wcpbnAbstractDb {

		public $wpdb;
		public $wpdb_prefix;
		public $table_name;
		public $mysql_version;
		public $plugin_prefix;
		public $default_special_keys = array();

		/**
		 * Construct.
		 *
		 * @return void
		 * @throws Exception
		 */
		public function __construct( $table_name ) {
			if ( ! WPO_WCPBN()->classes['database']->is_child_class_allowed( get_called_class() ) ) {
				throw new Exception( 'Unauthorized extension of ' . __CLASS__ . ' detected in ' . get_called_class() . '.' );
			}

			global $wpdb;
			$this->wpdb          = $wpdb;
			$this->wpdb_prefix   = $wpdb->prefix;
			$this->mysql_version = $this->wpdb->db_version();
			$this->plugin_prefix = 'wpo_wcpbn_';
			$this->table_name    = $table_name;
			
			$this->maybe_create_table();
		}

		public function get_full_table_name() {
			$db = apply_filters( 'wpo_wcpbn_alter_database_properties', $this );
			return $db->wpdb_prefix . $db->plugin_prefix . $db->table_name;
		}

		public function table_exists() {
			return $this->wpdb->get_var( "SHOW TABLES LIKE '{$this->get_full_table_name()}'" ) == $this->get_full_table_name();
		}
		
		public function maybe_create_table() {
			if ( ! $this->table_exists() ) {
				$this->create_table();
			}
		}
		
		protected function create_table() {}

		/**
		 * Get table rows.
		 * 
		 * @param array $args
		 * @param array $join_callback
		 *
		 * @return array|object|null
		 */
		public function get_rows( array $args, array $join_callback = array() ) {
			// parse dates
			foreach ( $this->get_date_keys() as $date ) {
				$args = ! empty( $args[ $date ] ) ? $this->parse_date_for_db_query( $args[ $date ], $date, $args ) : $args;
			}

			$sql         = "SELECT * FROM {$this->get_full_table_name()}";
			$sql_params  = array();
			$c           = 0;
			$next_clause = 'WHERE';

			// additional joins
			if ( ! empty( $join_callback ) && is_callable( $join_callback ) && WPO_WCPBN()->classes['database']->is_join_callback_allowed( $join_callback ) ) {
				$sql         = call_user_func_array( $join_callback, array( $sql, $args, $this->get_full_table_name() ) );
				$next_clause = 'AND';
			}

			foreach ( $args as $key => $value ) {
				if ( empty( $value ) || in_array( $key, $this->get_special_clauses_keys() ) ) {
					continue;
				}

				$key = sanitize_text_field( $key );

				// date args
				if ( 'date_query' === $key && is_array( $value ) ) {
					foreach ( $value as $date ) {
						$inclusive = ! empty( $date['inclusive'] ) ? '=' : '';
						if ( ! empty( $date['column'] ) ) {
							$date_column = sanitize_text_field( $date['column'] );
							$sql        .= " {$next_clause} {$date_column}";
							if ( ! empty( $date['after'] ) && ! empty( $date['before'] ) ) {
								$sql         .= " BETWEEN '%s' AND '%s'";
								$sql_params[] = $date['after'];
								$sql_params[] = $date['before'];
							} elseif ( ! empty( $date['after'] ) ) {
								$sql         .= " >{$inclusive} '%s'";
								$sql_params[] = $date['after'];
							} elseif ( ! empty( $date['before'] ) ) {
								$sql         .= " <{$inclusive} '%s'";
								$sql_params[] = $date['before'];
							} elseif ( ! empty( $date['equal'] ) ) {
								$sql         .= " = '%s'";
								$sql_params[] = $date['equal'];
							}
						}
					}
				// array args
				} elseif ( is_array( $value ) ) {
					$output          = [];
					$value           = array_walk_recursive( $value, function ( $v ) use ( &$output ) {
						$output[] = is_null( $v ) ? '' : (string) $v;
					} );
					$separated_comma = "('" . implode( "','", $output ) . "')";
					$sql            .= " {$next_clause} {$this->get_full_table_name()}.{$key} IN {$separated_comma}";
				// string args
				} elseif ( is_string( $value ) ) {
					$sql         .= " {$next_clause} {$this->get_full_table_name()}.{$key} = '%s'";
					$sql_params[] = $value;
				// integer args
				} elseif ( is_int( $value ) ) {
					$sql         .= " {$next_clause} {$this->get_full_table_name()}.{$key} = '%d'";
					$sql_params[] = $value;
				}

				$c++;
				if ( 1 === $c ) {
					$next_clause = 'AND';
				}
			}

			// prepare SQL query
			$sql = $this->prepare_sql_query( $sql, $sql_params );
			// special clauses
			$sql .= $this->get_special_clauses( $next_clause, $args );
			$sql .= $this->get_default_special_clauses( $next_clause, $args );

			return $this->wpdb->get_results( $sql );
		}

		/**
		 * Insert table row.
		 *
		 * @return int|false
		 */
		public function insert_row( $data ) {
			if ( is_array( $data ) ) {
				$query = $this->wpdb->insert(
					$this->get_full_table_name(),
					$data
				);
				if ( $query ) {
					return $this->wpdb->insert_id;
				} else {
					return false;
				}
			} else {
				return false;
			}
		}

		/**
		 * Update table row.
		 *
		 * @return int|false
		 */
		public function update_row( $data, $where ) {
			if ( is_array( $data ) && is_array( $where ) ) {
				return $this->wpdb->update(
					$this->get_full_table_name(),
					$data,
					$where
				);
			} else {
				return false;
			}
		}

		/**
		 * Update multiple rows of the table.
		 *
		 * Sample input:
		 * $data: array( row-id => array( 'some-column' => 'column_value' ), another-row-id => array( 'some-other-column' => 'column_value' ) )
		 *
		 * @param array<int, array<string, mixed>> $data
		 *
		 * @return bool
		 */
		public function update_rows( array $data ): bool {
			if ( empty( $data ) ) {
				return false;
			}

			$table_name = $this->get_full_table_name();
			$ids        = array();

			$columns = array_merge( ...array_values( $data ) );  // Get all unique columns across all rows.
			$columns = array_keys( $columns );  // Get the column names.

			$update_query = '';
			foreach ( $columns as $column ) {
				$column_update_query = '';
				foreach ( $data as $id => $update_data ) {
					if ( isset( $update_data[ $column ] ) ) {
						$ids[]               = $id;
						$column_update_query .= $this->wpdb->prepare( "WHEN %d THEN %s ", $id, $update_data[ $column ] );
					}
				}

				if ( ! empty ( $column_update_query ) ) {
					$update_query .= "{$column} = CASE id " . $column_update_query . "END, ";
				}
			}

			// Remove the trailing comma and space.
			$update_query = rtrim( $update_query, ', ' );

			if ( empty( $update_query ) ) {
				return false;
			}

			$ids = implode( ', ', array_unique( array_map( 'intval', $ids ) ) );

			$query  = "UPDATE {$table_name} SET " . $update_query . " WHERE id IN ({$ids});";
			$result = $this->wpdb->query( $query );

			return $result !== false;
		}

		/**
		 * Delete table row.
		 *
		 * @return int|false
		 */
		public function delete_row( $where ) {
			if ( is_array( $where ) ) {
				return $this->wpdb->delete(
					$this->get_full_table_name(),
					$where
				);
			} else {
				return false;
			}
		}

		/**
		 * Get date keys.
		 *
		 * @return array
		 */
		public function get_date_keys() {
			return array();
		}

		/**
		 * Get extended special clauses keys.
		 *
		 * @return array
		 */
		public function get_special_clauses_keys() {
			return $this->default_special_keys;
		}

		/**
		 * Get special clauses.
		 *
		 * @return string
		 */
		public function get_special_clauses( $next_clause, $args ) {
			return '';
		}

		/**
		 * Get default special clauses.
		 *
		 * @return string
		 */
		public function get_default_special_clauses( $next_clause, $args ) {
			$sql        = '';
			$sql_params = array();

			// search
			if ( ! empty( $args['search']['keys'] ) && is_array( $args['search']['keys'] ) && ! empty( $args['search']['term'] ) ) {
				if ( ! empty( $args['search']['type'] ) ) {
					switch ( $args['search']['type'] ) {
						case 'exact':
							$operator = '=';
							$term     = $args['search']['term'];
							break;
						case 'pattern':
							$operator = 'LIKE';
							$term     = "%{$args['search']['term']}%";
							break;
						case 'in':
							$operator = 'IN';
							if ( is_array( $args['search']['term'] ) && count( $args['search']['term'] ) ) {
								$term = '('.implode(',', $args['search']['term'] ).')';
							} else {
								$term = "({$args['search']['term']})";
							}
							break;
					}
				} else {
					$operator = '=';
					$term     = $args['search']['term'];
				}

				$search_clause = $next_clause;
				$c             = 0;
				foreach ( $args['search']['keys'] as $key ) {
					$key = sanitize_text_field( $key );
					
					if ( 'IN' === $operator ) {
						$sql         .= " {$search_clause} {$key} {$operator} {$term}";
					} else {
						$sql         .= " {$search_clause} {$key} {$operator} '%s'";
						$sql_params[] = $term;
					}

					$c++;
					if ( $c == 1 ) {
						$search_clause = 'OR';
					}
				}
			}

			// orderby/order
			if ( ! empty( $args['order_by'] ) && ! empty( $args['order'] ) ) {
				$order_by = sanitize_text_field( $args['order_by'] );
				$order    = sanitize_text_field( $args['order'] );
				$sql     .= " ORDER BY {$this->get_full_table_name()}.{$order_by} {$order}";
			}

			// limit/offset
			if ( ! empty( $args['limit'] ) && $args['limit'] > 0 ) {
				$sql         .= " LIMIT %d";
				$sql_params[] = $args['limit'];
				if ( ! empty( $args['offset'] ) && $args['offset'] > 0 ) {
					$sql         .= ", %d";
					$sql_params[] = $args['offset'];
				}
			}

			$sql = $this->prepare_sql_query( $sql, $sql_params );

			return $sql;
		}

		/**
		 * Map a valid date query var to DB arguments.
		 * Valid date formats: YYYY-MM-DD or timestamp, possibly combined with an operator from $valid_operators.
		 * Also accepts a WC_DateTime object.
		 * 
		 * Source: https://github.com/woocommerce/woocommerce/blob/3611d4643791bad87a0d3e6e73e031bb80447417/plugins/woocommerce/includes/data-stores/class-wc-data-store-wp.php#L334
		 *
		 * @since 3.2.0
		 * @param mixed  $query_var A valid date format.
		 * @param string $key meta or db column key.
		 * @param array  $query_args WP_Query args.
		 * @return array Modified $query_args
		 */
		public function parse_date_for_db_query( $query_var, $key, $query_args = array() ) {
			$query_parse_regex = '/([^.<>]*)(>=|<=|>|<|\.\.\.)([^.<>]+)/';
			$valid_operators   = array( '>', '>=', '=', '<=', '<', '...' );
	
			// YYYY-MM-DD queries have 'day' precision. Timestamp/WC_DateTime queries have 'second' precision.
			$precision = 'second';
	
			$dates    = array();
			$operator = '=';
	
			try {
				// Specific time query with a WC_DateTime.
				if ( is_a( $query_var, 'WC_DateTime' ) ) {
					$dates[] = $query_var;
				} elseif ( is_numeric( $query_var ) ) { // Specific time query with a timestamp.
					$dates[] = new WC_DateTime( "@{$query_var}", new DateTimeZone( 'UTC' ) );
				} elseif ( preg_match( $query_parse_regex, $query_var, $sections ) ) { // Query with operators and possible range of dates.
					if ( ! empty( $sections[1] ) ) {
						$dates[] = is_numeric( $sections[1] ) ? new WC_DateTime( "@{$sections[1]}", new DateTimeZone( 'UTC' ) ) : wc_string_to_datetime( $sections[1] );
					}
	
					$operator = in_array( $sections[2], $valid_operators, true ) ? $sections[2] : '';
					$dates[]  = is_numeric( $sections[3] ) ? new WC_DateTime( "@{$sections[3]}", new DateTimeZone( 'UTC' ) ) : wc_string_to_datetime( $sections[3] );
	
					if ( ! is_numeric( $sections[1] ) && ! is_numeric( $sections[3] ) ) {
						$precision = 'day';
					}
				} else { // Specific time query with a string.
					$dates[]   = wc_string_to_datetime( $query_var );
					$precision = 'day';
				}
			} catch ( Exception $e ) {
				return $query_args;
			}
	
			// Check for valid inputs.
			if ( ! $operator || empty( $dates ) || ( '...' === $operator && count( $dates ) < 2 ) ) {
				return $query_args;
			}
	
			// Build date query for date keys.
			if ( in_array( $key, $this->get_date_keys() ) ) {
				if ( ! isset( $query_args['date_query'] ) ) {
					$query_args['date_query'] = array();
				}
	
				$query_arg = array(
					'column'    => 'day' === $precision ? $key : $key,
					'inclusive' => '>' !== $operator && '<' !== $operator,
				);
	
				// Add 'after'/'before' query args.
				$comparisons = array();
				if ( '>' === $operator || '>=' === $operator || '...' === $operator ) {
					$comparisons[] = 'after';
				}
				if ( '<' === $operator || '<=' === $operator || '...' === $operator ) {
					$comparisons[] = 'before';
				}
	
				foreach ( $comparisons as $index => $comparison ) {
					if ( 'day' === $precision ) {
						$time = '00:00:00';
						if ( $comparison == 'before' ) {
							$time = '23:59:59';
						}
						$query_arg[ $comparison ] = gmdate( "Y-m-d {$time}", $dates[ $index ]->getTimestamp() );
					} else {
						$query_arg[ $comparison ] = gmdate( 'Y-m-d H:i:s', $dates[ $index ]->getTimestamp() );
					}
				}
	
				if ( empty( $comparisons ) ) {
					if ( strpos( $query_var, '-' ) === false ) {
						$query_arg['equal']  = gmdate( 'Y-m-d H:i:s', $dates[0]->getTimestamp() );
					} else {
						$query_arg['after']  = gmdate( 'Y-m-d 00:00:00', $dates[0]->getTimestamp() );
						$query_arg['before'] = gmdate( 'Y-m-d 23:59:59', $dates[0]->getTimestamp() );
					}
				}
	
				$query_args['date_query'][] = $query_arg;
				unset( $query_args[$key] );
			}
	
			return $query_args;
		}

		/**
		 * Prepares the SQL query string for safe execution.
		 * 
		 * @param  string  $sql
		 * @param  array   $sql_params
		 *
		 * @return string
		 */
		public function prepare_sql_query( $sql, $sql_params ) {
			if ( ! empty( $sql_params ) ) {
				return $this->wpdb->prepare( $sql, $sql_params );
			} else {
				return $sql;
			}
		}

		/**
		 * Catch database object errors.
		 * 
		 * @param  object  $wpdb
		 *
		 * @return array   Errors found
		 */
		public function catch_db_object_errors( $wpdb ) {
			global $EZSQL_ERROR;
		
			$errors = array();
		
			// using '$wpdb->queries'
			if ( ! empty( $wpdb->queries ) && is_array( $wpdb->queries ) ) {
				foreach ( $wpdb->queries as $query ) {
					$result = isset( $query['result'] ) ? $query['result'] : null;
		
					if ( is_wp_error( $result ) && is_array( $result->errors ) ) {
						foreach ( $result->errors as $error ) {
							$errors[] = reset( $error );
						}
					}
				}
			} 
		
			// fallback to '$EZSQL_ERROR'
			if ( empty( $errors ) && ! empty( $EZSQL_ERROR ) && is_array( $EZSQL_ERROR ) ) {
				foreach ( $EZSQL_ERROR as $error ) {
					$errors[] = $error['error_str'];
				}
			}
		
			// log errors
			if ( ! empty( $errors ) ) {
				foreach ( $errors as $error_message ) {
					WPO_WCPBN()->classes['logger']->log( $error_message, 'critical' );
				}
			}
		
			return $errors;
		}

		/**
		 * Alter database column data type.
		 *
		 * @return void
		 */
		public function alter_column_data_type( $column_name, $old_type, $new_type ) {
			// check if table exists
			if ( $this->table_exists() ) {
				// get the current data type
				$column_data_type = $this->wpdb->get_var( "SELECT DATA_TYPE FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME = '{$this->get_full_table_name()}' AND COLUMN_NAME = '{$column_name}'" );
				$old_type         = array( strtoupper( $old_type ), strtolower( $old_type ) );

				// check if the current data type is equal to the old type
				if ( ! empty( $column_data_type ) && in_array( $column_data_type, $old_type ) ) {
					$this->wpdb->query( "ALTER TABLE {$this->get_full_table_name()} MODIFY {$column_name} {$new_type}" );
				}
			}
		}

		/**
		 * Alter database datetime columns default to '1000-01-01 00:00:00'.
		 * 
		 * @param  array  $column_names
		 *
		 * @return void
		 */
		public function alter_datetime_columns_default( $column_names ) {
			if ( empty( $column_names ) || ! is_array( $column_names ) ) {
				return;
			}

			// check if table exists
			if ( $this->table_exists() ) {
				foreach ( $column_names as $column_name ) {
					// query
					$query = $this->wpdb->query( "ALTER TABLE {$this->get_full_table_name()} ALTER {$column_name} SET DEFAULT '1000-01-01 00:00:00'" );
					// log
					if ( $query ) {
						WPO_WCPBN()->classes['logger']->log( "Default value changed for '{$column_name}' column to '1000-01-01 00:00:00' on database table: {$this->get_full_table_name()}", 'info' );
					} else {
						WPO_WCPBN()->classes['logger']->log( "An error occurred! The default value for '{$column_name}' column couldn't be changed to '1000-01-01 00:00:00' on database table: {$this->get_full_table_name()}", 'critical' );
					}
				}
			}
		}

		/**
		 * Drops custom table.
		 * 
		 * @param  array  $table_name
		 *
		 * @return void
		 */
		private function drop_table( $table_name ) {
			// check if table exists
			if ( $this->wpdb->get_var( "SHOW TABLES LIKE '{$table_name}'" ) == $table_name ) {
				$this->wpdb->query( "DROP TABLE IF EXISTS {$table_name}" );
			}
		}

	}

}