<?php

namespace WPO\WC\Product_Batch_Numbers;

if ( ! defined( 'ABSPATH' ) ) {
	die;
}

if ( ! class_exists( 'wcpbnCsvExport' ) ) {

	class wcpbnCsvExport {

		protected static $_instance   = null;

		/**
		 * Instance
		 */
		public static function instance() {
			if ( is_null( self::$_instance ) ) {
				self::$_instance = new self();
			}
			return self::$_instance;
		}

		/**
		 * Construct
		 */
		public function __construct() {
			// ajax
			add_action( 'wp_ajax_wpo-wcpbn-export-customers', array( $this, 'export_csv_trigger' ) );
			add_action( 'wp_ajax_wpo-wcpbn-export-orders-items-batches', array( $this, 'export_csv_trigger' ) );
			add_action( 'wp_ajax_wpo-wcpbn-export-batches', array( $this, 'export_csv_trigger' ) );
			add_action( 'wp_ajax_wpo-wcpbn-export-order-batches', array( $this, 'export_csv_trigger' ) );
		}

		/**
		 * Catch AJAX export request
		 */
		public function export_csv_trigger() {
			// verify nonce
			check_ajax_referer( 'wpo_wcpbn_export', 'security' );

			// verify request
			if ( ! isset( $_POST ) || ! isset( $_POST['action'] ) ) {
				return;
			}

			// set data
			$data = array(
				'date_from' => ! empty( $_POST['from'] )   ? sanitize_text_field( $_POST['from'] )   : '',
				'date_to'   => ! empty( $_POST['to'] )     ? sanitize_text_field( $_POST['to'] )     : '',
				'status'    => ! empty( $_POST['status'] ) ? sanitize_text_field( $_POST['status'] ) : '',
				'wpai'      => isset( $_POST['wpai'] ) ? true : false,
			);

			if ( isset( $_POST['batch-id'] ) ) {
				$data['batch_id'] = sanitize_text_field( $_POST['batch-id'] );
			}

			if ( isset( $_POST['order-number'] ) ) {
				$data['order_id'] = sanitize_text_field( $_POST['order-number'] );
			}

			// export
			$this->export_csv( $data, $_POST['action'] );
		}

		/**
		 * Export a CSV file
		 */
		public function export_csv( $data, $action ) {
			if ( empty( $data ) || empty( $action ) ) {
				return;
			}

			$prefixes = array(
				'wpo-wcpbn-export-customers'            => 'woocommerce-product-batches-customers',
				'wpo-wcpbn-export-batches'              => 'woocommerce-product-batches',
				'wpo-wcpbn-export-orders-items-batches' => 'woocommerce-product-batches-order-items',
				'wpo-wcpbn-export-order-batches'		=> 'woocommerce-product-order-batch-numbers',
				'default'                               => 'woocommerce-product-batch-numbers',
			);
			$filename = sprintf( '%s-%s.csv', $prefixes[$action] ?? $prefixes['default'], date( 'Y-m-d_H-i' ) );

			header( "Content-Description: File Transfer" );
			header( "Content-Encoding: ".apply_filters( 'wpo_wcpbn_export_csv_header_content_encoding', 'UTF-8' ) );
			header( "Content-type: ".apply_filters( 'wpo_wcpbn_export_csv_header_content_type', 'text/csv; charset=UTF-8' ) );
			header( "Content-Disposition: attachment; filename=$filename" );
			$csv_data = $this->build_csv_data( $data, $action );
			$this->output_csv( $csv_data );
			die();
		}

		/**
		 * Output CSV
		 */
		public function output_csv( $data ) {
			$delimiter = apply_filters( 'wpo_wcpbn_export_csv_delimiter', "," );
			$enclosure = apply_filters( 'wpo_wcpbn_export_csv_enclosure', '"' );
			$escape    = apply_filters( 'wpo_wcpbn_export_csv_escape',    "\\" );
			$stream    = fopen( 'php://output', 'w' );
			// add BOM to fix UTF-8 in Excel: https://www.php.net/manual/de/function.fputcsv.php#121950
			fputs( $stream, chr(0xEF) . chr(0xBB) . chr(0xBF) );
			foreach ( $data as $row ) {
				fputcsv( $stream, $row, $delimiter, $enclosure, $escape );
			}
			fclose( $stream );
		}

		/**
		 * Build CSV data
		 */
		public function build_csv_data( $data, $action ) {
			if ( empty( $data ) || empty( $action ) ) {
				return array();
			}

			// csv data
			$csv_data = $this->get_export_data( $data, $action );
			if ( empty( $csv_data ) ) {
				return array();
			}

			// csv headers
			$first_row = reset( $csv_data );
			$headers   = array_keys( $first_row );

			// clean data rows
			foreach ( $csv_data as &$row ) {
				$row = array_values( array_map( 'htmlspecialchars_decode', $row ) );
			}

			// prepend headers
			array_unshift( $csv_data, $headers );

			return $csv_data;
		}

		/**
		 * Get export data
		 */
		public function get_export_data( $data, $action ) {
			if ( empty( $data ) || empty( $action ) ) {
				return;
			}
			
			$export_data = array();
			
			switch ( str_replace( 'wpo-wcpbn-export-', '', $action ) ) {
				case "customers":
					$export_data = $this->get_order_customers_by_batch( $data );
					break;
				case "orders-items-batches":
					$export_data = $this->get_orders_items_batches( $data );
					break;
				case "batches":
					$export_data = $this->get_batches( $data );
					break;
				case "order-batches":
					$export_data = $this->get_order_batches( $data );
					break;
			}

			// allow additional data to be added
			return apply_filters( 'wpo_wcpbn_export_csv_data', $export_data, $action );
		}

		/**
		 * Get order customers by batch
		 */
		public function get_order_customers_by_batch( $data ) {
			$batch_order_ids    = null;
			$export             = array();
			$context            = 'order-customers';
			$valid_order_status = apply_filters( 'wpo_wcpbn_export_csv_order_status', array( 'wc-completed' ), $context );

			if ( isset( $data['batch_id'] ) ) {
				$batch_order_ids = wpo_wcpbn_get_batch_order_ids( $data['batch_id'] );
			} else {
				WPO_WCPBN()->classes['logger']->log( 'No batch ID given for export.' );
				return $export;
			}
		
			if ( ! empty( $batch_order_ids ) ) {
				foreach ( $batch_order_ids as $order_id ) {
					$order = wc_get_order( $order_id );
					
					if ( empty( $order ) ) {
						WPO_WCPBN()->classes['logger']->log( "Export order is empty for the ID #{$order_id}." );
						continue;
					}
					
					if ( ! in_array( "wc-{$order->get_status()}", $valid_order_status ) ) {
						WPO_WCPBN()->classes['logger']->log( "Export order status not valid for the ID #{$order_id}." );
						continue;
					}
					
					$customer_id                          = $order->get_customer_id();
		
					$export[ $customer_id ]['id']         = $customer_id;
					$export[ $customer_id ]['first_name'] = $order->get_billing_first_name();
					$export[ $customer_id ]['last_name']  = $order->get_billing_last_name();
					$export[ $customer_id ]['company']    = $order->get_billing_company();
					$export[ $customer_id ]['email']      = $order->get_billing_email();
					
					$export[ $customer_id ]               = apply_filters( 'wpo_wcpbn_export_csv_customer_data', $export[ $customer_id ], $customer_id, $order );
				}
		
				return array_values( apply_filters( 'wpo_wcpbn_export_csv_customers_data', $export, $batch_order_ids ) );
			} else {
				WPO_WCPBN()->classes['logger']->log( "No export orders found for the given batch ID #{$data['batch_id']}." );
				return $export;
			}
		}

		/**
		 * Get order items batches
		 */
		public function get_orders_items_batches( $data ) {
			$export = [];
			$order  = [];

			// date-range
			if ( ! empty( $data['date_from'] ) && ! empty( $data['date_to'] ) ) {
				$context = 'order-items';
				$orders  = wc_get_orders( array(
					'limit'        => -1,
					'type'         => 'shop_order',
					'status'       => apply_filters( 'wpo_wcpbn_export_csv_order_status', array( 'wc-completed' ), $context ),
					'date_created' => $data['date_from'] .'...'. $data['date_to'] 
				) );
		
				if ( empty( $orders ) ) {
					WPO_WCPBN()->classes['logger']->log( 'No export orders found for the date range provided.' );
				}
		
			// order-number
			} elseif ( ! empty( $data['order_id'] ) ) {
				$orders[] = wc_get_order( $data['order_id'] );
				if ( empty( $orders ) ) {
					WPO_WCPBN()->classes['logger']->log( 'Order not found.' );
				}
			} else {
				WPO_WCPBN()->classes['logger']->log( 'Data for export is missing' );
			}
			
			if ( ! empty( $orders ) ) {
				foreach ( $orders as $order ) {
					if ( empty( $order ) ) {
						WPO_WCPBN()->classes['logger']->log( 'Export order object is empty.' );
						continue;
					}
		
					$hook = apply_filters( 'wpo_wcpbn_export_csv_orders_items_batches_before_order_data', true, $order );
					if ( ! $hook ) {
						continue;
					}
		
					foreach ( $order->get_items() as $item_id => $item ) {
						$batch_meta = $item->get_meta( '_wcpbn_data' );
						if ( empty( $batch_meta ) || ! is_array( $batch_meta ) ) {
							continue;
						}
						$product_id = $item['variation_id'] != 0 ? $item['variation_id'] : $item['product_id'];
						$product    = wc_get_product( $product_id );
						$sku        = ! empty( $product ) && is_callable( array( $product, 'get_sku' ) ) ? $product->get_sku() : '';
		
						foreach ( $batch_meta as $batch ) {
							$export[$item_id.'_'.$batch['id']]['order_id']          = $order->get_id();
							$export[$item_id.'_'.$batch['id']]['order_item_id']     = $item_id;
							$export[$item_id.'_'.$batch['id']]['sku']               = $sku;
							$shared_ids = wpo_wcpbn_get_batch_shared_product_ids( $batch['id'] );
							$export[$item_id.'_'.$batch['id']]['shared_ids']        = ! empty( $shared_ids ) ? implode( '|', $shared_ids ) : '';
							$export[$item_id.'_'.$batch['id']]['qty']               = $item->get_quantity();
							$export[$item_id.'_'.$batch['id']]['batch_id']          = $batch['id'];
							$export[$item_id.'_'.$batch['id']]['batch_number']      = $batch['batch_number'];
							$export[$item_id.'_'.$batch['id']]['batch_qty_reduced'] = $batch['quantity_sold'];
						}
					}
				}
			}
			
			return array_values( apply_filters( 'wpo_wcpbn_export_csv_orders_items_batches_data', $export, $orders ) );
		}

		/**
		 * Get batches data
		 */
		public function get_batches( $data ) {
			$export = array();

			if ( isset( $data['date_from'] ) && isset( $data['date_to'] ) && isset( $data['status'] ) ) {
				$batches = wpo_wcpbn_get_batches_between_date_range( $data['date_from'], $data['date_to'], $data['status'] );
		
				if ( empty( $batches ) ) {
					WPO_WCPBN()->classes['logger']->log( 'No matching batches found for export.' );
					return $export;
				}
		
				foreach ( $batches as $batch ) {
					$row = array();
					
					if ( $data['wpai'] ) {
						$row['batch_id'] = $batch->id;
					} else {
						$row['id'] = $batch->id;
					}
					
					if ( ! $data['wpai'] ) {
						$row['date_created'] = $batch->date_created;
					}
					
					$row['date_expiry'] = $batch->date_expiry;
					
					if ( ! $data['wpai'] ) {
						$row['user_id'] = $batch->user_id;
					}
					
					$row['batch_number'] = $batch->batch_number;
					$row['supplier']     = $batch->supplier;
					
					if ( wc_string_to_bool( WPO_WCPBN()->classes['settings']->options->get_option( 'stock_locations' ) ) ) {
						$row['location'] = isset( $batch->location ) ? $batch->location : '';
					}
					
					$row['quantity']           = wpo_wcpbn_force_stock_amount_type( $batch->quantity );
					$row['quantity_available'] = wpo_wcpbn_force_stock_amount_type( $batch->quantity_available );
					$row['status']             = $batch->status;
					
					if ( ! $data['wpai'] ) {
						$row['order_ids'] = substr( $batch->order_ids, 1, -1 );
					}
					
					$shared_ids = wpo_wcpbn_get_batch_shared_product_ids( $batch->id );
					
					if ( ! empty( $shared_ids ) ) {
						if ( $data['wpai'] ) {
							foreach ( $shared_ids as $product_id ) {
								$row['product_id'] = $product_id;
								$export[]          = $row;
							}
							continue;
						} else {
							$row['shared_ids'] = ! empty( $shared_ids ) ? implode( '|', $shared_ids ) : '';
						}
					}
					
					$export[] = $row;
				}
		
				return array_values( apply_filters( 'wpo_wcpbn_export_csv_batches_data', $export, $batches ) );
		
			} else {
				WPO_WCPBN()->classes['logger']->log( 'Data for export is missing' );
				return $export;
			}
		}

		/**
		 *	Get batches by order id. 
		 */
		public function get_order_batches( $data ) {
			$export = array();
			if ( ! empty( $data['order_id'] ) ) {
				$batches = WPO_WCPBN()->classes['database']->batch_numbers()->get_batches_by_order_id( $data['order_id'] );

				if ( empty( $batches ) ) {
					WPO_WCPBN()->classes['logger']->log( 'No matching batches found for export.' );
					return $export;
				}
		
				foreach ( $batches as $batch ) {
				
					$export[$batch->id]['id'] 			= $batch->id;	
					$export[$batch->id]['date_expiry'] 	= $batch->date_expiry;	
					$export[$batch->id]['batch_number'] = $batch->batch_number;
					$export[$batch->id]['supplier'] 	= $batch->supplier;
					if ( wc_string_to_bool( WPO_WCPBN()->classes['settings']->options->get_option( 'stock_locations' ) ) ) {
						$export[$batch->id]['location'] = isset( $batch->location ) ? $batch->location : '';
					}
					$export[$batch->id]['quantity']    		  = wpo_wcpbn_force_stock_amount_type( $batch->quantity );
					$export[$batch->id]['quantity_available'] = wpo_wcpbn_force_stock_amount_type( $batch->quantity_available );
					$export[$batch->id]['status']             = $batch->status;					
				}
		
				return array_values( apply_filters( 'wpo_wcpbn_export_csv_batches_data', $export, $batches ) );
			} else {
				WPO_WCPBN()->classes['logger']->log( 'Data for export is missing' );
				return $export;
			}		
		}

	}

}