<?php

namespace WPO\WC\Product_Batch_Numbers;

use WPO\WC\Product_Batch_Numbers\wcpbnSemaphore as Semaphore;

if ( ! defined( 'ABSPATH' ) ) {
	die;
}

if ( ! class_exists( 'wcpbnInstall' ) ) {

	class wcpbnInstall {
		
		public $lock_name;
		public $lock_context;
		public $lock_time;
		public $lock_retries;
		protected static $_instance = null;

		/**
		 * Class instance
		 */
		public static function instance() {
			if ( is_null( self::$_instance ) ) {
				self::$_instance = new self();
			}
			return self::$_instance;
		}

		/**
		 * Construct
		 */
		public function __construct() {
			$this->lock_name    = 'wpo_wcpbn_install_lock';
			$this->lock_context = array( 'source' => 'wpo-wcpbn-install' );
			$this->lock_time    = apply_filters( 'wpo_wcpbn_install_lock_time', 2 );
			$this->lock_retries = apply_filters( 'wpo_wcpbn_install_lock_retries', 0 );
		
			// run lifecycle methods
			if ( is_admin() && ! defined( 'DOING_AJAX' ) ) {
				add_action( 'wp_loaded', array( $this, 'do_install' ) );
			}

			add_action( 'admin_notices', array( $this, 'legacy_export_addon_enabled_notice' ), 1 );
			add_action( 'admin_notices', array( $this, 'legacy_supplier_to_be_deleted_notice' ), 1 );
			add_action( 'wpo_wcpbn_migrate_product_ids_1_9_0', array( $this, 'add_product_id_to_batch_1_9_0' ), 10, 2 );
		}

		/**
		 * Handles version checking
		 */
		public function do_install() {
			$installed_version = get_option( WPO_WCPBN()->version_setting );

			// installed version lower than plugin version?
			if ( version_compare( $installed_version, WPO_WCPBN()->version, '<' ) ) {
				if ( ! $installed_version ) {
					$this->install();
				} else {
					$this->upgrade( $installed_version );
				}
			} else {
				$this->downgrade( $installed_version );
			}
		}

		/**
		 * Plugin install method. Perform any installation tasks here
		 */
		protected function install() {
			// Update version setting
			update_option( WPO_WCPBN()->version_setting, WPO_WCPBN()->version );
			// Show batch number in order on activation
			WPO_WCPBN()->classes['settings']->options->update_option( 'display_batch_number', 'yes' );
			// Set the global date expiry offset value to 30 days
			WPO_WCPBN()->classes['settings']->options->update_option( 'global_expiration_offset', 30 );
		}

		/**
		 * Plugin upgrade method. Perform any required upgrades here
		 *
		 * @param  string  $installed_version  The currently installed ('old') version
		 */
		protected function upgrade( $installed_version ) {
			if ( version_compare( $installed_version, '1.0.1', '<' ) ) {
				// Migrates database column 'order_ids' data type from 'json' to 'longtext' to support legacy MySQL versions
				WPO_WCPBN()->classes['database']->batch_numbers()->alter_column_data_type( 'order_ids', 'json', 'longtext' );
			}

			if ( version_compare( $installed_version, '1.2.5', '<' ) ) {
				// Migrates database columns data type from 'smallint' to 'int' to support legacy MySQL versions
				WPO_WCPBN()->classes['database']->batch_numbers()->alter_column_data_type( 'quantity', 'smallint', 'int' );
				WPO_WCPBN()->classes['database']->batch_numbers()->alter_column_data_type( 'quantity_available', 'smallint', 'int' );
			}

			if ( version_compare( $installed_version, '1.2.9-dev-1', '<' ) ) {
				// Migrates database datetime columns default to '1000-01-01 00:00:00'
				WPO_WCPBN()->classes['database']->batch_numbers()->alter_datetime_columns_default( array( 'date_created', 'date_expiry' ) );
			}

			if ( version_compare( $installed_version, '1.3.0-dev-1', '<' ) ) {
				// Migrates database columns data type from 'smallint' or 'int' to decimal(10,4) to support float numbers
				WPO_WCPBN()->classes['database']->batch_numbers()->alter_column_data_type( 'quantity', 'smallint', 'decimal(10,4)' );
				WPO_WCPBN()->classes['database']->batch_numbers()->alter_column_data_type( 'quantity_available', 'smallint', 'decimal(10,4)' );
				WPO_WCPBN()->classes['database']->batch_numbers()->alter_column_data_type( 'quantity', 'int', 'decimal(10,4)' );
				WPO_WCPBN()->classes['database']->batch_numbers()->alter_column_data_type( 'quantity_available', 'int', 'decimal(10,4)' );
			}

			if ( version_compare( $installed_version, '1.3.1-dev-1', '<' ) ) {
				$this->migrate_legacy_settings_1_3_0();
			}

			if ( version_compare( $installed_version, '1.3.3-dev-1', '<' ) ) {
				if ( function_exists( 'deactivate_plugins' ) ) {
					deactivate_plugins( 'woocommerce-product-batch-numbers-export/woocommerce-product-batch-numbers-export.php' );
				}
			}

			if ( version_compare( $installed_version, '1.4.2-dev-1', '<' ) ) {
				$this->migrate_settings_1_4_1();
			}

			if ( version_compare( $installed_version, '1.7.1-dev-1', '<' ) ) {
				// we need to delete a custom table created without prefixes by mistake on version 1.7.0, plugin will resume using the correct table
				$table_name = 'batch_numbers';
				if ( WPO_WCPBN()->classes['database']->batch_numbers()->wpdb->get_var( "SHOW TABLES LIKE '{$table_name}'" ) == $table_name ) {
					WPO_WCPBN()->classes['database']->batch_numbers()->wpdb->query( "DROP TABLE {$table_name}" );
				}

				// clean archive batches pending actions which were created in high volume
				if ( function_exists( 'as_unschedule_all_actions' ) ) {
					as_unschedule_all_actions( 'wcpbn_schedule_archive_batches' );
				}
			}

			if ( version_compare( $installed_version, '1.8.0-dev-1', '<' ) ) {
				// adds the location database column for existing table
				$table_name = WPO_WCPBN()->classes['database']->batch_numbers()->get_full_table_name();
				WPO_WCPBN()->classes['database']->batch_numbers()->wpdb->query( "ALTER TABLE {$table_name} ADD location varchar(200) DEFAULT '' AFTER supplier" );
			}
			
			if ( version_compare( $installed_version, '1.9.0-dev-2', '<' ) ) {				
				// migrate products to the new shared products table
				// we are not removing the 'product_id' and 'variation_id' columns from the batch table now
				// we should do it in a later version
				$this->migrate_product_ids_1_9_0();
				
				// adds the units database column for existing table
				$table_name = WPO_WCPBN()->classes['database']->batch_numbers()->get_full_table_name();
				WPO_WCPBN()->classes['database']->batch_numbers()->wpdb->query( "ALTER TABLE {$table_name} ADD units smallint(5) DEFAULT 1 AFTER quantity_available" );
			}
			
			if ( version_compare( $installed_version, '3.0.4-beta-1', '<' ) ) {
				// Migrates database columns data type to 'decimal(20,4)' to support bigger float numbers
				$table_name = WPO_WCPBN()->classes['database']->batch_numbers()->get_full_table_name();
				WPO_WCPBN()->classes['database']->batch_numbers()->wpdb->query( "ALTER TABLE {$table_name} MODIFY quantity decimal(20,4) NOT NULL" );
				WPO_WCPBN()->classes['database']->batch_numbers()->wpdb->query( "ALTER TABLE {$table_name} MODIFY quantity_available decimal(20,4) NOT NULL" );
			}

			// Update settings keys to new tabbed settings keys.
			if ( version_compare( $installed_version, '3.1.0-beta-2', '<' ) ) {
				$this->migrate_settings_3_0_4();
			}

			if ( version_compare( $installed_version, '3.1.2-beta-1', '<' ) ) {
				$this->migrate_settings_3_1_1();
			}

			// Update version setting
			update_option( WPO_WCPBN()->version_setting, WPO_WCPBN()->version );
		}

		/**
		 * Plugin downgrade method. Perform any required downgrades here
		 * 
		 * @param  string  $installed_version  The currently installed ('old') version (actually higher since this is a downgrade)
		 */
		protected function downgrade( $installed_version ) {
			// Update version setting
			update_option( WPO_WCPBN()->version_setting, WPO_WCPBN()->version );
		}

		/**
		 * Migrates legacy settings <= v1.3.0 to the new ones
		 */
		public function migrate_legacy_settings_1_3_0() {
			$options_matching = array(
				'wcpbn_show_date_expiry_order' => array(
					'checkbox' => true,
					'option'   => 'display_expiry_date',
				),
				'wcpbn_show_batch_number_order' => array(
					'checkbox' => true,
					'option'   => 'display_batch_number',
				),
				'wcpbn_date_expiry_global_offset_warning' => array(
					'checkbox' => false,
					'option'   => 'global_expiration_offset',
				),
				'wcpbn_date_expiry_offset_email_notification' => array(
					'checkbox' => true,
					'option'   => 'expiration_offset_notification',
				),
				'wcpbn_date_expiry_offset_notify_every_day' => array(
					'checkbox' => true,
					'option'   => 'expiration_offset_notification_every_day',
				),
				'wcpbn_include_batch_numbers_in_formatted_item_meta' => array(
					'checkbox' => true,
					'option'   => 'include_number_in_formatted_item_meta',
				),
				'wcpbn_include_batch_expiry_dates_in_formatted_item_meta' => array(
					'checkbox' => true,
					'option'   => 'include_expiration_date_in_formatted_item_meta',
				),
				'wcpbn_auto_archive_expired_batches' => array(
					'checkbox' => true,
					'option'   => 'archive_expired',
				),
				'wcpbn_assign_expired_batches_to_order_items' => array(
					'checkbox' => true,
					'option'   => 'assign_expired',
				),
				'wcpbn_show_batches_wc_products_listing_column' => array(
					'checkbox' => true,
					'option'   => 'display_batch_column_products',
				),
				'wcpbn_show_batches_wc_orders_listing_column' => array(
					'checkbox' => true,
					'option'   => 'display_batch_column_orders',
				),
				'wcpbn_batch_stock_reduction_priority' => array(
					'checkbox' => false,
					'option'   => 'batch_prioritization',
				),
				'wcpbn_sync_with_wc_stock' => array(
					'checkbox' => true,
					'option'   => 'sync_wc_stock',
				),
				'wcpbn_delete_batches_on_product_deletion' => array(
					'checkbox' => true,
					'option'   => 'delete_batch_on_product_delete',
				),
				'wcpbn_stock_threshold_email_notifications' => array(
					'checkbox' => true,
					'option'   => 'sold_out_notification',
				),
				'wcpbn_batch_expired_email_notification' => array(
					'checkbox' => true,
					'option'   => 'expiration_notification',
				),
				'wcpbn_batch_notifications_email_address' => array(
					'checkbox' => false,
					'option'   => 'email_address_type',
				),
				'wcpbn_batch_notifications_email_address__custom' => array(
					'checkbox' => false,
					'option'   => 'custom_email_address',
				),
			);

			foreach ( $options_matching as $legacy => $new ) {
				$legacy_value = get_option( $legacy );
				if ( $new['checkbox'] ) {
					$legacy_value = ( $legacy_value == 1 ) ? 'yes' : 'no';
				}
				if ( $legacy_value ) {
					WPO_WCPBN()->classes['settings']->options->update_option( $new['option'], $legacy_value );
				}
				delete_option( $legacy );
			}
		}

		public function legacy_export_addon_enabled_notice() {
			if ( is_plugin_active( 'woocommerce-product-batch-numbers-export/woocommerce-product-batch-numbers-export.php' ) ) {
				?>
				<div class="error">
					<p>
					<?php
						printf(
							/* translators: here */
							__( 'Your legacy <strong>WooCommerce Product Batch Numbers Export</strong> add-on is still enabled. You can safely uninstall it from %s.', 'woocommerce-product-batch-numbers' ),
							'<a href="'.esc_url_raw( network_admin_url( 'plugins.php?s=WooCommerce+Product+Batch+Numbers+Export+Add-on' ) ).'">'.__( 'here', 'woocommerce-product-batch-numbers' ).'</a>'
						);
					?>
					</p>
				</div>
				<?php
			}
		}

		public function migrate_settings_1_4_1() {
			$legacy_settings = array(
				'archive_expired' => array(
					'option'        => 'archive_batches',
					'values'        => array(
						'yes' => 'expired',
						'no'  => 'no',
					),
				),
				'expiring_offset_notification' => array(
					'option'        => 'expiration_date_offset_notification',
				),
				'display_batch_number' => array(
					'option'        => 'display_batch_data_in_order_admin',
					'values'        => array(
						'yes' => wc_string_to_bool( WPO_WCPBN()->classes['settings']->options->get_option( 'display_expiry_date' ) ) ? 'number+date' : 'number',
						'no'  => 'no',
					),
				),
				'display_expiry_date' => array(
					'option'        => 'display_batch_data_in_order_admin',
					'values'        => array(
						'yes' => wc_string_to_bool( WPO_WCPBN()->classes['settings']->options->get_option( 'display_batch_number' ) ) ? 'number+date' : 'date',
						'no'  => 'no',
					),
				),
				'include_number_in_formatted_item_meta' => array(
					'option'        => 'display_batch_data_in_formatted_item_meta',
					'values'        => array(
						'yes' => wc_string_to_bool( WPO_WCPBN()->classes['settings']->options->get_option( 'include_expiration_date_in_formatted_item_meta' ) ) ? 'number+date' : 'number',
						'no'  => 'no',
					),
				),
				'display_batch_data_in_formatted_item_meta' => array(
					'option'        => 'display_batch_data_in_formatted_item_meta',
					'values'        => array(
						'yes' => wc_string_to_bool( WPO_WCPBN()->classes['settings']->options->get_option( 'include_number_in_formatted_item_meta' ) ) ? 'number+date' : 'date',
						'no'  => 'no',
					),
				),
				'display_batch_column_products' => array(
					'option'        => 'display_batch_column_products',
					'values'        => array(
						'yes' => wc_string_to_bool( WPO_WCPBN()->classes['settings']->options->get_option( 'display_expiration_date_in_batch_column_products' ) ) ? 'number+date' : 'number',
						'no'  => 'no',
					),
				),
				'display_batch_column_orders' => array(
					'option'        => 'display_batch_column_orders',
					'values'        => array(
						'yes' => 'number',
						'no'  => 'no',
					),
				),
			);

			foreach ( $legacy_settings as $legacy_setting => $new ) {
				$legacy_value = WPO_WCPBN()->classes['settings']->options->get_option( $legacy_setting );
				$new_value    = ! empty( $new['values'][$legacy_value] ) ? $new['values'][$legacy_value] : $legacy_value;
				WPO_WCPBN()->classes['settings']->options->update_option( $new['option'], $new_value );
				if ( $legacy_setting != $new['option'] ) {
					WPO_WCPBN()->classes['settings']->options->delete_option( $legacy_setting );
				}
			}
		}
		
		public function migrate_product_ids_1_9_0() {
			if ( ! function_exists( 'as_schedule_single_action' ) ) {
				return;
			}
		
			$lock    = new Semaphore( $this->lock_name, $this->lock_time, array( wc_get_logger() ), $this->lock_context );
			$batches = wpo_wcpbn_get_batches();
			
			if ( ! empty( $batches ) ) {
				foreach ( $batches as $batch ) {
					$product_id = ! empty( $batch->variation_id ) ? $batch->variation_id : $batch->product_id;
					$hook       = 'wpo_wcpbn_migrate_product_ids_1_9_0';
					
					if ( $lock->lock( $this->lock_retries ) ) {
						try {
							as_schedule_single_action( time(), $hook, [ 'batch_id' => $batch->id, 'product_id' => $product_id ] );
						} catch ( \Exception $e ) {
							$lock->log( $e, 'critical' );
						} catch ( \Error $e ) {
							$lock->log( $e, 'critical' );
						}
			
						$lock->release();
					} else {
						$lock->log( "Couldn't get the lock!", 'critical' );
					}
				}
			}
		}
		
		public function add_product_id_to_batch_1_9_0( $batch_id, $product_id ) {
			$query = wpo_wcpbn_add_products_to_batch( $batch->id, [ intval( $product_id ) ] );
			if ( ! $query ) {
				WPO_WCPBN()->classes['logger']->log(
					sprintf( 'An error occurred while migrating the product #%1$s from the batch #%2$s to the new shared products table.' ),
					esc_attr( $product_id ),
					esc_attr( $batch_id )
				);
			}
		}
		
		public function legacy_supplier_to_be_deleted_notice() {
			if ( ! is_super_admin() ) {
				return;
			}
			
			if ( ! get_option( 'wpo_wcpbn_hide_legacy_supplier_field_notice' ) ) {
				ob_start();
				?>
				<div class="notice notice-warning">
					<p><?php _e( 'The WooCommerce Product Batch Numbers plugin will remove the legacy Supplier batch field from the next version. Please use the new Batch Purchase Order Supplier field in replacement.', 'woocommerce-product-batch-numbers' ); ?></p>
					<p><a href="<?php echo esc_url( wp_nonce_url( add_query_arg( 'wpo_wcpbn_hide_legacy_supplier_field_notice', 'true' ), 'hide_legacy_supplier_field_notice_nonce' ) ); ?>"><?php _e( 'Hide this message', 'woocommerce-product-batch-numbers' ); ?></a></p>
				</div>
				<?php
				echo wp_kses_post( ob_get_clean() );
			}
			
			// save option to hide notice
			if ( isset( $_REQUEST['wpo_wcpbn_hide_legacy_supplier_field_notice'] ) && isset( $_REQUEST['_wpnonce'] ) ) {
				// validate nonce
				if ( ! wp_verify_nonce( $_REQUEST['_wpnonce'], 'hide_legacy_supplier_field_notice_nonce' ) ) {
					wcpdf_log_error( 'You do not have sufficient permissions to perform this action: wpo_wcpbn_hide_legacy_supplier_field_notice' );
					wp_redirect( 'edit.php?post_type=product&page=wpo-batch-numbers' );
					exit;
				} else {
					update_option( 'wpo_wcpbn_hide_legacy_supplier_field_notice', true );
					wp_redirect( 'edit.php?post_type=product&page=wpo-batch-numbers' );
					exit;
				}
			}
		}

		public function migrate_settings_3_0_4() {
			$legacy_settings = get_option( 'wpo_wcpbn_settings_options' );
			if ( ! empty( $legacy_settings ) ) {
				$general_settings = array(
					'batch_prioritization'             => $legacy_settings['batch_prioritization'] ?? '',
					'global_expiration_offset'         => $legacy_settings['global_expiration_offset'] ?? '',
					'archive_batches'                  => $legacy_settings['archive_batches'] ?? '',
					'assign_expired'                   => $legacy_settings['assign_expired'] ?? '',
					'sync_wc_stock'                    => $legacy_settings['sync_wc_stock'] ?? '',
					'stock_locations'                  => $legacy_settings['stock_locations'] ?? '',
					'delete_batch_on_product_delete'   => $legacy_settings['delete_batch_on_product_delete'] ?? '',
					'custom_date_format'               => $legacy_settings['custom_date_format'] ?? '',
					'batch_date_label'                 => $legacy_settings['batch_date_label'] ?? '',
				);

				$po_settings = array(
					'use_po_prices'                   => $legacy_settings['use_po_prices'] ?? '',
					'po_price_prioritization'         => $legacy_settings['po_price_prioritization'] ?? '',
					'limit_cart_qty_to_keep_po_price' => $legacy_settings['limit_add_to_cart_qty_to_keep_po_price'] ?? '',
				);

				$display_settings = array(
					'display_batch_data_in_order_admin'               => $legacy_settings['display_batch_data_in_order_admin'] ?? '',
					'display_batch_data_in_cart'                      => $legacy_settings['display_batch_data_in_cart'] ?? '',
					'display_batch_data_in_product_pages'             => $legacy_settings['display_batch_data_in_product_pages'] ?? '',
					'display_batch_data_in_product_pages_location'    => $legacy_settings['display_batch_data_in_product_pages_location'] ?? '',
					'display_batch_data_in_product_pages_limit'       => $legacy_settings['display_batch_data_in_product_pages_limit'] ?? '',
					'display_batch_data_in_product_pages_time_window' => $legacy_settings['display_batch_data_in_product_pages_time_window'] ?? '',
					'display_batch_data_in_formatted_item_meta'       => $legacy_settings['display_batch_data_in_formatted_item_meta'] ?? '',
					'display_batch_column_products'                   => $legacy_settings['display_batch_column_products'] ?? '',
					'display_batch_column_orders'                     => $legacy_settings['display_batch_column_orders'] ?? '',
				);

				$email_settings = array(
					'expiration_offset_notification'           => $legacy_settings['expiration_offset_notification'] ?? '',
					'expiration_offset_notification_every_day' => $legacy_settings['expiration_offset_notification_every_day'] ?? '',
					'sold_out_notification'                    => $legacy_settings['sold_out_notification'] ?? '',
					'expiration_notification'                  => $legacy_settings['expiration_notification'] ?? '',
					'archived_on_offset_notification'          => $legacy_settings['archived_on_offset_notification'] ?? '',
					'expiration_date_offset_notification'      => $legacy_settings['expiration_date_offset_notification'] ?? '',
					'email_address_type'                       => $legacy_settings['email_address_type'] ?? '',
					'custom_email_address'                     => $legacy_settings['custom_email_address'] ?? '',
				);

				$advanced_settings = array(
					'log_batch_deletion'               => $legacy_settings['log_batch_deletion'] ?? '',
					'batch_units'                      => $legacy_settings['batch_units'] ?? '',
					'delete_duplicate_shared_products' => $legacy_settings['delete_duplicate_shared_products'] ?? '',
				);

				update_option( 'wpo_wcpbn_settings_options_general', $general_settings );
				update_option( 'wpo_wcpbn_settings_options_po', $po_settings );
				update_option( 'wpo_wcpbn_settings_options_display', $display_settings );
				update_option( 'wpo_wcpbn_settings_options_email', $email_settings );
				update_option( 'wpo_wcpbn_settings_options_advanced', $advanced_settings );

				delete_option( 'wpo_wcpbn_settings_options' );
			}
		}

		public function migrate_settings_3_1_1() {
			$general_settings = get_option( 'wpo_wcpbn_settings_options_general', array() );
			$general_settings = array(
				'batch_prioritization'           => $general_settings['batch_prioritization'] ?? '',
				'global_expiration_offset'       => $general_settings['global_expiration_offset'] ?? '',
				'archive_batches'                => $general_settings['archive_batches'] ?? '',
				'assign_expired'                 => $general_settings['assign_expired'] ?? '',
				'sync_wc_stock'                  => $general_settings['sync_wc_stock'] ?? '',
				'stock_locations'                => $general_settings['stock_locations'] ?? '',
				'delete_batch_on_product_delete' => $general_settings['delete_batch_on_product_delete'] ?? '',
				'custom_date_format'             => $general_settings['custom_date_format'] ?? '',
				'batch_date_label'               => $general_settings['batch_date_label'] ?? '',
			);
			update_option( 'wpo_wcpbn_settings_options_general', $general_settings );

			$po_settings = get_option( 'wpo_wcpbn_settings_options_po', array() );
			$po_settings = array(
				'use_po_prices'                   => $po_settings['use_po_prices'] ?? '',
				'po_price_prioritization'         => $po_settings['po_price_prioritization'] ?? '',
				'limit_cart_qty_to_keep_po_price' => $po_settings['limit_cart_qty_to_keep_po_price'] ?? '',
			);
			update_option( 'wpo_wcpbn_settings_options_po', $po_settings );

			$display_settings = get_option( 'wpo_wcpbn_settings_options_display', array() );
			$display_settings = array(
				'display_batch_data_in_order_admin'               => $display_settings['display_batch_data_in_order_admin'] ?? '',
				'display_batch_data_in_cart'                      => $display_settings['display_batch_data_in_cart'] ?? '',
				'display_batch_data_in_product_pages'             => $display_settings['display_batch_data_in_product_pages'] ?? '',
				'display_batch_data_in_product_pages_location'    => $display_settings['display_batch_data_in_product_pages_location'] ?? '',
				'display_batch_data_in_product_pages_limit'       => $display_settings['display_batch_data_in_product_pages_limit'] ?? '',
				'display_batch_data_in_product_pages_time_window' => $display_settings['display_batch_data_in_product_pages_time_window'] ?? '',
				'display_batch_data_in_formatted_item_meta'       => $display_settings['display_batch_data_in_formatted_item_meta'] ?? '',
				'display_batch_column_products'                   => $display_settings['display_batch_column_products'] ?? '',
				'display_batch_column_orders'                     => $display_settings['display_batch_column_orders'] ?? '',
			);
			update_option( 'wpo_wcpbn_settings_options_display', $display_settings );

			$email_settings = get_option( 'wpo_wcpbn_settings_options_email', array() );
			$email_settings = array(
				'expiration_offset_notification'           => $email_settings['expiration_offset_notification'] ?? '',
				'expiration_offset_notification_every_day' => $email_settings['expiration_offset_notification_every_day'] ?? '',
				'sold_out_notification'                    => $email_settings['sold_out_notification'] ?? '',
				'expiration_notification'                  => $email_settings['expiration_notification'] ?? '',
				'archived_on_offset_notification'          => $email_settings['archived_on_offset_notification'] ?? '',
				'expiration_date_offset_notification'      => $email_settings['expiration_date_offset_notification'] ?? '',
				'email_address_type'                       => $email_settings['email_address_type'] ?? '',
				'custom_email_address'                     => $email_settings['custom_email_address'] ?? '',
			);
			update_option( 'wpo_wcpbn_settings_options_email', $email_settings );

			$advanced_settings = get_option( 'wpo_wcpbn_settings_options_advanced', array() );
			$advanced_settings = array(
				'log_batch_deletion'               => $advanced_settings['log_batch_deletion'] ?? '',
				'batch_units'                      => $advanced_settings['batch_units'] ?? '',
				'delete_duplicate_shared_products' => $advanced_settings['delete_duplicate_shared_products'] ?? '',
			);
			update_option( 'wpo_wcpbn_settings_options_advanced', $advanced_settings );
		}

	}

}