<?php
/**
 * Order Items Class
 *
 * @since 1.0.0
 */

namespace WPO\WC\Product_Batch_Numbers;

use WC_Order_Item_Product;

/**
 * If this file is called directly, abort.
 *
 * @since 1.0.0
 */
if ( ! defined( 'ABSPATH' ) ) {
	die;
}

if ( ! class_exists( 'wcpbnOrderItems' ) ) {

	class wcpbnOrderItems {

		protected        $wc_manage_stock;
		protected        $orderby;
		protected        $allowed_stock_actions;
		protected static $_instance = null;

		/**
		 * Instance
		 */
		public static function instance() {
			if ( is_null( self::$_instance ) ) {
				self::$_instance = new self();
			}
			return self::$_instance;
		}

		/**
		 * Construct
		 */
		public function __construct() {
			add_action( 'woocommerce_admin_order_item_headers', array( $this, 'order_items_headers' ), 10, 1 );
			add_action( 'woocommerce_admin_order_item_values', array( $this, 'order_item_values' ), 10, 3 );
			add_action( 'woocommerce_before_save_order_items', array( $this, 'on_save_order_items' ), 10, 2 );
			add_action( 'woocommerce_before_delete_order_item', array( $this, 'on_delete_order_item' ), 10, 2 );
			if ( version_compare( WC_VERSION, '7.1', '<' ) ) {
				add_action( 'wp_trash_post', array( $this, 'restore_stock' ), 10, 1 );
			} else {
				add_action( 'woocommerce_trash_order', array( $this, 'restore_stock' ), 10, 1 ); // WC 7.1+
			}
			if ( version_compare( WC_VERSION, '7.2', '<' ) ) {
				add_action( 'untrashed_post', array( $this, 'reduce_stock' ), 10, 1 );	
			} else {
				add_action( 'woocommerce_untrash_order', array( $this, 'reduce_stock' ), 10, 1 ); // WC 7.2+
			}
			add_filter( 'woocommerce_hidden_order_itemmeta', array( $this, 'hide_batch_itemmeta_wc_order' ), 10, 1 );

			// formatted item meta
			if ( ! empty( WPO_WCPBN()->classes['settings']->options->get_option( 'display_batch_data_in_formatted_item_meta' ) ) && WPO_WCPBN()->classes['settings']->options->get_option( 'display_batch_data_in_formatted_item_meta' ) != 'no' ) {
				add_filter( 'woocommerce_order_item_get_formatted_meta_data', array( $this, 'include_in_formatted_item_meta' ), 10, 2 );
			}

			$this->wc_manage_stock = get_option( 'woocommerce_manage_stock' );
			
			// prevents unwanted stock changes
			$this->allowed_stock_actions = apply_filters( 'wpo_wcpbn_order_items_allowed_stock_actions', [
				'reduce' => [
					'woocommerce_reduce_order_stock',
					'yith_pos_reduce_order_stock',
					'rgw_payment_processed',
					'wcs_renewal_order_created',
					'woocommerce_payment_complete',
					'woocommerce_order_status_completed',
					'woocommerce_order_status_processing',
					'woocommerce_order_status_on-hold',
				],
				'restore' => [
					'woocommerce_restore_order_stock',
					'yith_pos_restore_order_stock',
					'woocommerce_order_status_cancelled',
					'woocommerce_order_status_pending',
					'woocommerce_refund_created',
				],
			], $this );

			if ( $this->wc_manage_stock === 'yes' ) {
				// RedSys Gateway for WooCommerce Pro (by Codection): https://github.com/wpovernight/woocommerce-product-batch-numbers/pull/256/commits/6439aac8f4f4c73801e9240ec89457418aaadba2
				if ( ! is_admin() && class_exists( 'WC_Redsys_Gateway' ) ) { // Prevents duplicated stock changes
					add_action( 'rgw_payment_processed', array( $this, 'rgw_reduce_stock' ), 99, 3 );

				// defaults
				} else {
					// default
					add_action( 'woocommerce_reduce_order_stock', array( $this, 'reduce_stock' ), 10, 1 );
					add_action( 'woocommerce_restore_order_stock', array( $this, 'restore_stock' ), 10, 1 );
					
					// YITH POS
					add_action( 'yith_pos_reduce_order_stock', array( $this, 'reduce_stock' ), 10, 1 );
					add_action( 'yith_pos_restore_order_stock', array( $this, 'restore_stock' ), 10, 1 );

					// used by WooCommerce Subscriptions on renewal orders
					add_filter( 'wcs_renewal_order_created', array( $this, 'wcs_reduce_stock' ), 10, 2  );
				}
				
			} else {
				// On maybe reduce stock levels
				add_action( 'woocommerce_payment_complete', array( $this, 'reduce_stock' ), 10, 1 );
				add_action( 'woocommerce_order_status_completed', array( $this, 'reduce_stock' ), 10, 1 );
				add_action( 'woocommerce_order_status_processing', array( $this, 'reduce_stock' ), 10, 1 );
				add_action( 'woocommerce_order_status_on-hold', array( $this, 'reduce_stock' ), 10, 1 );
				// On maybe restore stock levels
				add_action( 'woocommerce_order_status_cancelled', array( $this, 'restore_stock' ), 10, 1 );
				add_action( 'woocommerce_order_status_pending', array( $this, 'restore_stock' ), 10, 1 );
			}

			// restore stock on refunds
			add_action( 'woocommerce_refund_created', array( $this, 'restore_stock_on_refund' ), 10, 2 );
			
			// get batch order by priority
			$this->orderby = ( $batch_stock_reduction_priority = WPO_WCPBN()->classes['settings']->options->get_option( 'batch_prioritization' ) ) ? $batch_stock_reduction_priority : 'id';
		}

		/**
		 * Adds order items table header columns.
		 *
		 * @return void
		 */
		public function order_items_headers( $order ) {
			$setting = WPO_WCPBN()->classes['settings']->options->get_option( 'display_batch_data_in_order_admin' );
			if ( in_array( $setting, array( 'date', 'number+date' ) ) ) {
				echo '<th style="min-width:100px;">'.__( 'Batch Expiry', 'woocommerce-product-batch-numbers' ).'</th>';
			}
			if ( in_array( $setting, array( 'number', 'number+date' ) ) ) {
				echo '<th style="min-width:100px;">'.__( 'Batch Numbers', 'woocommerce-product-batch-numbers' ).'</th>';
			}
		}

		/**
		 * Adds batch number and expiry date data to order items table columns.
		 *
		 * @return void
		 */
		public function order_item_values( $product, $item, $item_id ) {
			$setting              = WPO_WCPBN()->classes['settings']->options->get_option( 'display_batch_data_in_order_admin' );
			$display_batch_date   = in_array( $setting, array( 'date', 'number+date' ) );
			$display_batch_number = in_array( $setting, array( 'number', 'number+date' ) );

			if ( is_object( $product ) && ! empty( $item ) ) {
				$product_id = wpo_wcpbn_wpml_object_id( $product->get_id() ); // gets the original product ID if it's a language copy
				$product    = wc_get_product( $product_id );

				if ( empty( $product ) ) {
					return;
				}

				// Get active batch numbers for this product
				$query_args = array(
					'shared_ids' => [ intval( $product_id ) ],
					'status'     => 'active',
				);
				$product_active_batch_numbers = wpo_wcpbn_get_batches( $query_args );

				// Get item meta
				$item_batch_numbers_data = wc_get_order_item_meta( $item_id, '_wcpbn_data', true );

				if ( ! empty( $item_batch_numbers_data ) ) {

					// Expiry date column data
					if ( $display_batch_date ) {
						echo '<td class="wpo-wcpbn-batch-expire-dates-column">';
						foreach ( $item_batch_numbers_data as $item_batch_number ) {
							$results = wpo_wcpbn_get_batches( array( 'id' => $item_batch_number['id'] ) );
							if( ! empty( $results[0]->date_expiry ) && $results[0]->date_expiry != '0000-00-00 00:00:00' && $results[0]->date_expiry != '1000-01-01 00:00:00' ) {
								echo '<p style="margin-top:0px;">'.date( wpo_wcpbn_date_format(), strtotime( $results[0]->date_expiry ) ).'</p>';
							} else {
								echo '<p style="margin-top:0px;">'.__('n/a', 'woocommerce-product-batch-numbers').'</p>';
							}
						}
						echo '</td>';
					}

					// Batch number column data
					if ( $display_batch_number ) {
						echo '<td class="wpo-wcpbn-batch-numbers-column">';
						foreach ( $item_batch_numbers_data as $item_batch_number ) {
							echo '<p style="margin-top:0px;">';
							echo $item_batch_number['batch_number'].'<span class="wpo-wcpbn-order-item-quantity">'.$item_batch_number['quantity_sold'].'</span>';
							echo '</p>';
							echo '<p class="wcpbn-item-batch-number-edit" style="display:none;">';
							echo '<select style="padding:4px; width:60%;" name="wcpbn-update-batch-number['.$item_id.'][]">';
							echo '<option value="'.$item_batch_number['id'].'" selected>'.$item_batch_number['batch_number'].'</option>';
							foreach ( $product_active_batch_numbers as $active_batch_number ) {
								if ( $item_batch_number['id'] != $active_batch_number->id ) {
									echo '<option value="'.$active_batch_number->id.'">'.$active_batch_number->batch_number.'</option>';
								}
							}
							echo '</select>';
							echo '<input type="hidden" name="wcpbn-batch-number-id['.$item_id.'][]" value="'.$item_batch_number['id'].'">';
							$qty_pattern = has_filter( 'woocommerce_stock_amount', 'intval' ) ? '[0-9]*' : '';
							echo '<input style="width:30%; text-align:center; vertical-align:middle;" type="number" class="wcpbn-update-batch-number-quantity-sold" name="wcpbn-update-batch-number-quantity-sold['.$item_id.'][]" value="'.$item_batch_number['quantity_sold'].'" min="0" step="any" pattern="'.$qty_pattern.'">';
							echo '</p>';
						}
						echo '</td>';
					}
				} else {
					// Expiry date column data
					if ( $display_batch_date ) {
						echo '<td class="wpo-wcpbn-batch-expire-dates-column">'.__('n/a', 'woocommerce-product-batch-numbers').'</td>';
					}
					// Batch number column data
					if ( $display_batch_number ) {
						echo '<td class="wpo-wcpbn-batch-numbers-column">'.__('n/a', 'woocommerce-product-batch-numbers').'</td>';
					}
				}

			} else { // For shipping and others
				if ( $display_batch_date ) {
					echo '<td></td>';
				}
				if ( $display_batch_number ) {
					echo '<td></td>';
				}
			}
		}

		/**
		 * Allocates the batch numbers and quantities to apply to each order item.
		 *
		 * @return array
		 */
		private function allocated_item_batch_numbers( $product_id, $item, $item_quantity, $product_type, $order_id ) {
			// batch units
			if ( wc_string_to_bool( WPO_WCPBN()->classes['settings']->options->get_option( 'batch_units' ) ) ) {
				
			}
			$item_quantity = apply_filters( 'wpo_wcpbn_order_item_quantity_for_batch_allocation', $item_quantity, $item, $order_id );
			$response      = $this->allocate_batch_numbers( $product_id, $item_quantity, $item, $order_id );

			return apply_filters( 'wpo_wcpbn_order_item_batch_number', $response, $item );
		}

		/**
		 * Allocates the batch numbers for a given product/item.
		 *
		 * @return array
		 */
		public function allocate_batch_numbers( $product_id, $quantity, $item = null, $order_id = null ) {
			$response = array();

			// WPML compatibility
			$product_id = wpo_wcpbn_wpml_object_id( $product_id ); // gets the original product ID if it's a language copy

			// mirror enabled?
			$is_mirroring = wpo_wcpbn_is_product_mirroring_batches_stock( $product_id );

			$query_args = array(
				'shared_ids' => [ intval( $product_id ) ],
				'status'     => 'active',
				'order_by'   => $this->orderby,
				'order'      => 'ASC',
			);
			$db_rows = wpo_wcpbn_get_batches( $query_args );

			if ( ! empty( $db_rows ) ) {

				// hook before item batch allocation
				$db_rows = apply_filters( 'wpo_wcpbn_before_allocate_batch_numbers', $db_rows, $product_id, $item, $order_id );
				$db_rows = apply_filters( 'wpo_wcpbn_before_allocate_order_item_batch_numbers', $db_rows, $product_id, $item, $order_id ); // legacy filter

				// if assign expired batches to order items is disabled we need to remove those from the array
				if ( ! wc_string_to_bool( WPO_WCPBN()->classes['settings']->options->get_option( 'assign_expired' ) ) ) {
					$today = time();
					foreach ( $db_rows as $key => $batch ) {
						if ( $batch->date_expiry != '0000-00-00 00:00:00' && $batch->date_expiry != '1000-01-01 00:00:00' ) { // when date expiry is not set on the database ignore
							$batch_date = strtotime( $batch->date_expiry );
							if ( $today > $batch_date ) {
								unset( $db_rows[$key] );
							}
						}
					}
					// reorder array keys after the unset
					$db_rows = array_values( $db_rows );
				}

				// default allocation
				if ( ! $is_mirroring ) {

					// keep track of the remaining quantity
					$remaining_qty = wpo_wcpbn_force_stock_amount_type( $quantity );

					// map stock to batches
					foreach ( $db_rows as $key => $batch ) {
						$quantity_available = wpo_wcpbn_force_stock_amount_type( $batch->quantity_available );
						$units              = 1;
						
						// batch units enabled
						if ( wc_string_to_bool( WPO_WCPBN()->classes['settings']->options->get_option( 'batch_units' ) ) ) {
							$units = ! empty( $batch->units ) ? intval( $batch->units ) : $units;
							if ( $units > 1 ) {
								$remaining_qty *= $units;
							}
						}

						// not enough space
						if ( $quantity_available < $units ) {
							continue;
						}

						// add to allocation response
						$response[$batch->id]                  = get_object_vars( $batch );
						$response[$batch->id]['quantity_sold'] = $remaining_qty - ( max( 0, $remaining_qty - $quantity_available ) );

						// subtract remaining to allocate
						$remaining_qty -= $response[$batch->id]['quantity_sold'];

						// no need to keep going if nothing to allocate
						if ( $remaining_qty <= 0 ) {
							break;
						}
					}
				
				// mirror allocation
				} else {
					// map batches
					foreach ( $db_rows as $key => $batch ) {
						$quantity_available = wpo_wcpbn_force_stock_amount_type( $batch->quantity_available );

						// not enough space
						if ( $quantity_available < 1 ) {
							continue;
						}

						// add to allocation response
						$response[$batch->id]                  = get_object_vars( $batch );
						$response[$batch->id]['quantity_sold'] = wpo_wcpbn_force_stock_amount_type( $quantity );
					}
				}

			}

			return apply_filters( 'wpo_wcpbn_allocated_batch_numbers', $response, $product_id );
		}

		/**
		 * Increments the batch number quantity sold on WC reduce stock.
		 *
		 * @return void
		 */
		public function reduce_stock( $order ) {
			// some actions provide the order_id directly others only the order object
			if ( is_object( $order ) ) {
				$order_id = $order->get_id();
			} else {
				$order_id = $order;
			}

			$this->on_save_order_items( $order_id, null );
		}

		/**
		 * Decrements the batch number quantity sold on WC restore stock.
		 *
		 * @return void
		 */
		public function restore_stock( $order ) {
			// some actions provide the order_id directly others only the order object
			if ( is_object( $order ) ) {
				$order_id = $order->get_id();
			} else {
				$order_id = $order;
			}
			
			$this->on_save_order_items( $order_id, null );
		}

		/**
		 * Decrements the batch number quantity sold on WC restore stock on refund.
		 *
		 * @return void
		 */
		public function restore_stock_on_refund( $refund_id, $args ) {
			$this->on_save_order_items( $refund_id, null );
		}

		/**
		 * Updates order items meta and database values on save.
		 *
		 * @return void
		 */
		public function on_save_order_items( $order_id, $items ) {
			$wc_reduce_restore = false;
			// reduce or restore stock
			if ( is_null( $items ) ) {
				$wc_reduce_restore = true;
			}

			if ( empty( $items ) ) {
				if ( is_object( $order_id ) ) {
					$order    = $order_id;
					$order_id = $order->get_id();
				} else {
					$order    = wc_get_order($order_id);
				}
				
				if ( ! empty( $order ) ) {
					$items = $order->get_items();
				}
			} elseif ( ! empty( $items ) && ! empty( $items['order_item_id'] ) ) {
				// save items args to grab input values
				$items_post_args = $items; // not array of item objects
				$items_arr       = array();
				$order           = wc_get_order($order_id);
				foreach ( $items['order_item_id'] as $item_id ) {
					$items_arr[$item_id] = $order->get_item( $item_id );
				}
				$items = $items_arr; // array of item objects
			} else {
				return;
			}

			if ( ! is_array( $items ) ) {
				return;
			}

			foreach ( $items as $item_id => $item ) {
				// Old meta
				$old_item_meta = wc_get_order_item_meta( $item_id, '_wcpbn_data' );

				// Editing batch numbers and quantities in order item and update batch quantities only if the order status is allowed
				if ( isset( $items_post_args['wcpbn-batch-number-id'][$item_id] ) && ! empty( $items_post_args['wcpbn-batch-number-id'][$item_id] ) && in_array( $order->get_status(), $this->get_allowed_stock_order_status( 'reduce' ) ) ) {

					if ( ! empty( $old_item_meta ) ) {

						// Change Array keys to numeric (0, 1, 2, ...)
						$item_meta_arr = array_values( $old_item_meta );

						// prepare data
						$data = array();

						// Form input values
						for ( $i = 0; $i < count( $item_meta_arr ); $i++ ) {
							$data[$i]['id']            = $items_post_args['wcpbn-update-batch-number'][$item_id][$i];
							$data[$i]['quantity_sold'] = wpo_wcpbn_force_stock_amount_type( $items_post_args['wcpbn-update-batch-number-quantity-sold'][$item_id][$i] );
							// get batch number because the form select uses the ID
							$db_row                    = wpo_wcpbn_get_batches( array( 'id' => $data[$i]['id'] ) );
							$data[$i]['batch_number']  = $db_row[0]->batch_number;

							// if batch number is different 
							if ( $item_meta_arr[$i]['id'] != $data[$i]['id'] ) {

								// check if the order stock was reduced
								if ( ! empty( $order ) && $order->get_meta( '_wcpbn_order_stock_reduced' ) == 'yes' ) {
									// Restore old item row before updating the new one
									$this->db_row_save( $item_meta_arr[$i], 'restore', $order_id, 'all_items' );
								}
								// Update the new one
								$this->db_row_save( $data[$i], 'reduce', $order_id, 'all_items' );
								// unset the old batch
								unset( $old_item_meta[$item_meta_arr[$i]['id']] );
								// rename the array key to use the batch ID instead or default numeric
								foreach ( $data as $key => $value ) {
									$data[$data[$key]['id']] = $value;
									unset( $data[$key] );
								}

								// Update item meta merging the new data
								wc_update_order_item_meta( $item_id, '_wcpbn_data', $this->clean_order_item_batch_for_meta( $old_item_meta + $data ) );
							
							// if both batches are the same
							} elseif ( $item_meta_arr[$i]['id'] == $data[$i]['id'] ) {

								// Check if the quantity sold has changed
								if ( $item_meta_arr[$i]['quantity_sold'] != $data[$i]['quantity_sold'] ) {

									// check if the new quantity is superior
									if ( $item_meta_arr[$i]['quantity_sold'] < $data[$i]['quantity_sold'] ) {
										// get the remaining quantity
										$qty_remaining = $item_meta_arr[$i]['quantity_sold'] = $data[$i]['quantity_sold'] - $item_meta_arr[$i]['quantity_sold'];
										// reduce remaining from item row
										$this->db_row_save( $item_meta_arr[$i], 'reduce', $order_id, 'all_items' );
										// re-add the remaining to update meta
										$old_item_meta[$item_meta_arr[$i]['id']]['quantity_sold'] = $old_item_meta[$item_meta_arr[$i]['id']]['quantity_sold'] + $qty_remaining;
										// update item meta
										wc_update_order_item_meta( $item_id, '_wcpbn_data', $this->clean_order_item_batch_for_meta($old_item_meta) );
										
									// check if the new quantity is inferior
									} elseif ( $item_meta_arr[$i]['quantity_sold'] > $data[$i]['quantity_sold'] ) {
										// if new quantity is zero
										if ( $data[$i]['quantity_sold'] == 0 ) {
											// restore old item row sold quantity
											$this->db_row_save( $item_meta_arr[$i], 'restore', $order_id, 'all_items' );
											// unset the batch from the item meta to update
											unset( $old_item_meta[$data[$i]['id']] );
											// update item meta with the new quantity
											wc_update_order_item_meta( $item_id, '_wcpbn_data', $this->clean_order_item_batch_for_meta($old_item_meta) );
										} else {
											// grab the quantity difference
											$item_meta_arr[$i]['quantity_sold'] = $item_meta_arr[$i]['quantity_sold'] - $data[$i]['quantity_sold'];
											// restore the difference on the item
											$this->db_row_save( $item_meta_arr[$i], 'restore', $order_id, 'all_items' );
											// Set the new quantity to update the meta
											$old_item_meta[$item_meta_arr[$i]['id']]['quantity_sold'] = $data[$i]['quantity_sold'];
											// update item meta width the new quantity
											wc_update_order_item_meta( $item_id, '_wcpbn_data', $this->clean_order_item_batch_for_meta($old_item_meta) );
										}
									}

								} else {
									continue;
								}
							} else {
								continue;
							}

						}

					}

				// Saving/updating batches on order item when not reducing/restoring stock
				} else {
					if ( empty( $item ) || ! is_object( $item ) || $item->get_type() != 'line_item' ) { // if we don't have a item continue to next
						continue;
					}

					$item_quantity = $item->get_quantity();
					$product       = $item->get_product(); // Get the product data from item
					
					if ( empty( $product ) ) {
						continue;
					}

					// WPML compatibility
					$product_id = wpo_wcpbn_wpml_object_id( $product->get_id() ); // gets the original product ID if it's a language copy
					$product    = apply_filters( 'wpo_wcpbn_order_item_product_object', wc_get_product( $product_id ), $item );

					if ( empty( $product ) ) {
						continue;
					}

					// Don't fire this if we are reducing stock (for example when the order comes in)
					if ( ! $wc_reduce_restore && in_array( $order->get_status(), $this->get_allowed_stock_order_status( 'reduce' ) ) ) {

						// Check if this item has already batch saved meta
						if ( ! empty( $old_item_meta ) ) {
							$saved_quantity_sold = array();
							foreach ( $old_item_meta as $key => $row ) {
								$saved_quantity_sold[] = $row['quantity_sold'];
							}
							if ( array_sum( $saved_quantity_sold ) < $item_quantity ) {
								$item_quantity = $item_quantity - array_sum( $saved_quantity_sold );
							} else {
								continue; // quantities are the same, nothing to do on this item
							}
						}

						// Query for available batch numbers for this item
						$batch_numbers_of_item = $this->allocated_item_batch_numbers( $product_id, $item, $item_quantity, $product->get_type(), $order_id );
						if ( ! empty( $batch_numbers_of_item ) ) {
							// reduce on database
							foreach ( $batch_numbers_of_item as $key => $row ) {
								// Update the new one
								$this->db_row_save( $row, 'reduce', $order_id, 'all_items' );

								// update item meta if exists
								if ( ! empty( $old_item_meta ) ) {
									$batch_numbers_of_item[$key]['quantity_sold'] = $old_item_meta[$key]['quantity_sold'] + $item_quantity;
								}
							}
							// update item meta
							wc_update_order_item_meta( $item_id, '_wcpbn_data', $this->clean_order_item_batch_for_meta( $batch_numbers_of_item ) );
						}
					}
				}

			}

			// On reduce or restore stock actions
			if ( $wc_reduce_restore ) {
				$this->reduce_and_restore_stock_actions( $order_id );
			}

			// save batch single meta (to be used by third party exporters like Advanced Order Export For WooCommerce)
			$this->save_order_items_single_batch_meta( $order_id );

		}

		/**
		 * Reduce or restores stock levels on a given order ID by checking the action.
		 *
		 * @return void
		 */
		private function reduce_and_restore_stock_actions( $order_id ) {
			// current action
			$current_action = current_action();

			if ( ! ( $order = wc_get_order( $order_id ) ) ) {
				return;
			}

			$meta_updated = false;

			// Get order stock reduced meta
			if ( empty( $stock_reduced = $order->get_meta( '_wcpbn_order_stock_reduced' ) ) ) {
				$stock_reduced = 'no';
				$order->update_meta_data( '_wcpbn_order_stock_reduced', $stock_reduced );
				$meta_updated = true;
			}

			// check if we are dealing with a refund
			if ( $order->get_type() == 'shop_order_refund' ) {
				$parent_order_id = $order->get_parent_id();
				$order           = wc_get_order( $parent_order_id );
				$order_id        = $order->get_id();
				$is_refund       = true;
				$restore         = $_POST['restock_refunded_items'] == 'true' ? true : false;
			} else {
				$is_refund       = false;
				$restore         = true;
			}

			foreach ( $order->get_items() as $item_id => $item ) {
				$product_id = $item['variation_id'] != 0 ? $item['variation_id'] : $item['product_id'];
				$product_id = wpo_wcpbn_wpml_object_id( $product_id ); // gets the original product ID if it's a language copy
				$product    = apply_filters( 'wpo_wcpbn_order_item_product_object', wc_get_product( $product_id ), $item );

				if ( empty( $product ) ) {
					continue;
				}

				$product_id   = $product->get_id();
				$product_type = $product->get_type();

				// if it's a item reduce
				if ( in_array( $current_action, $this->allowed_stock_actions['reduce'] ) ) {
					$trigger_reduce = $stock_reduced == 'yes' || empty( $stock_reduced ) ?? 'no';

					// Only continue if we're reducing stock & WC Stock Management is disabled.
					if ( $this->wc_manage_stock == 'no' ) {
						if ( $trigger_reduce == 'no' ) {
							continue;
						}
					}

					// Process reduce
					if ( in_array( $order->get_status(), $this->get_allowed_stock_order_status( 'reduce' ) ) ) {
						// Query for available batch numbers for this item
						$batch_numbers_of_item = $this->allocated_item_batch_numbers( $product_id, $item, $item->get_quantity(), $product_type, $order_id );

						if ( ! empty( $batch_numbers_of_item ) ) {
							foreach ( $batch_numbers_of_item as $row ) {
								// Get the row to update quantity sold
								$this->db_row_save( $row, 'reduce', $order_id, 'all_items' );
							}
							
							// update item meta
							wc_update_order_item_meta( $item_id, '_wcpbn_data', $this->clean_order_item_batch_for_meta( $batch_numbers_of_item ) );
						}
						// Ensure stock is marked as "reduced" in case payment complete or other stock actions are called.
						$order->update_meta_data( '_wcpbn_order_stock_reduced', 'yes' );
						$meta_updated = true;
					} else {
						// Ensure stock is not marked as "reduced" anymore.
						$order->update_meta_data( '_wcpbn_order_stock_reduced', 'no' );
						$meta_updated = true;
					}
				}

				// if it's a item restore
				if ( in_array( $current_action, $this->allowed_stock_actions['restore'] ) ) {
					$trigger_increase = $stock_reduced;

					// Only continue if we're increasing stock & WC Stock Management is disabled.
					if ( $this->wc_manage_stock == 'no' ) {
						if ( $trigger_increase == 'no' ) {
							continue;
						}
					}

					// If trash is trigger replace the allowed status
					if ( in_array( current_action(), [ 'wp_trash_post', 'woocommerce_trash_order' ] ) ) {
						$allowed_status = $this->get_allowed_stock_order_status('reduce');
					} else {
						$allowed_status = $this->get_allowed_stock_order_status('restore');
					}

					// Process restore
					if ( in_array( $order->get_status(), $allowed_status ) || $is_refund ) { // refunds can have parent orders from any status
						if ( ! empty( $item_meta = $item->get_meta( '_wcpbn_data' ) ) ) {
							// we have a refund
							if ( $is_refund ) {
								$line_item_qtys  = json_decode( sanitize_text_field( stripslashes( $_POST['line_item_qtys'] ) ), true );
								$item_refund_qty = isset( $line_item_qtys[ $item_id ] ) ? wpo_wcpbn_force_stock_amount_type( $line_item_qtys[ $item_id ] ) : null ;

								if ( $restore && $item_refund_qty > 0 ) { // only if the checkbox 'Restock refunded items' is enabled
									$refund_rows = array();
									foreach ( $item_meta as $key => $row ) {
										if ( $row['quantity_sold'] >= $item_refund_qty ) {
											$refund_rows[$key]                  = $row;
											$refund_rows[$key]['quantity_sold'] = $item_refund_qty;
											break;
										} else {
											$refund_rows[$key] = $row;
											$item_refund_qty  -= $row['quantity_sold'];
										}
									}
									$item_meta = $refund_rows;
								} else {
									continue;
								}
							}
							
							// restore
							if ( $restore ) {
								foreach ( $item_meta as $row ) {
									// Get the row to update quantity sold
									$this->db_row_save( $row, 'restore', $order_id, 'all_items' );
								}

								// Ensure stock is not marked as "reduced" anymore.
								$order->update_meta_data( '_wcpbn_order_stock_reduced', 'no' );
								$meta_updated = true;
							}
						}
					}
				}
			}

			if ( $meta_updated ) {
				$order->save_meta_data();
			}
		}

		/**
		 * Update batch numbers DB table on order item deletion.
		 *
		 * @return void
		 */
		public function on_delete_order_item( $item_id ) {
			$item     = new WC_Order_Item_Product( $item_id );
			$order_id = $item->get_order_id();

			if ( empty( $item_id ) || empty( $order_id ) || empty( $item ) ) {
				return;
			}

			if ( ! empty( $item_meta = $item->get_meta( '_wcpbn_data' ) ) ) {
				// Get order object
				$order = wc_get_order( $order_id );

				if ( empty( $order ) ) {
					return;
				}

				// Update batch quantities only if the order status is 'completed'
				if ( in_array($order->get_status(), $this->get_allowed_stock_order_status( 'reduce' ) ) ) {
					foreach ( $item_meta as $row ) {
						// Process row
						$this->db_row_save( $row, 'restore', $order_id, 'solo_item' );
					}
				}
				// Delete item meta
				wc_delete_order_item_meta( $item_id, '_wcpbn_data' );
			}
		}

		/**
		 * Gets DB row to process and update.
		 *
		 * @return void
		 */
		private function db_row_save( $row, $operation_type, $order_id, $movement_type = null ) {
			if ( $this->data_operation_types( $operation_type ) == false ) {
				return;
			}

			if ( ! in_array( $operation_type, array( 'reduce', 'update', 'restore' ) ) ) {
				return;
			}
			
			$batch = wpo_wcpbn_get_batches( [ 'id' => $row['id'] ] );
			if ( empty( $batch ) ) {
				return;
			}
			$batch = is_array( $batch ) ? reset( $batch ) : $batch;

			if ( ! empty( $batch ) ) {
				// Set quantity sold
				if ( $operation_type == 'reduce' ) {
					$data['quantity_available'] = wpo_wcpbn_force_stock_amount_type( $batch->quantity_available ) - wpo_wcpbn_force_stock_amount_type( $row['quantity_sold'] );
				} elseif ( in_array( $operation_type, array( 'update', 'restore' ) ) ) {
					$data['quantity_available'] = wpo_wcpbn_force_stock_amount_type( $batch->quantity_available ) + wpo_wcpbn_force_stock_amount_type( $row['quantity_sold'] );
				}

				if ( in_array( $operation_type, array( 'reduce', 'update' ) ) ) {
					// Set status archived
					if ( $data['quantity_available'] == 0 ) {
						$data['status'] = 'archived';

						// schedule email notification
						if ( wc_string_to_bool( WPO_WCPBN()->classes['settings']->options->get_option( 'sold_out_notification' ) ) ) {
							WPO_WCPBN()->classes['batches']->schedule_stock_notification_email( $batch );
						}
					}

					// Set the order ID
					if ( ! empty( $batch->order_ids ) ) {
						$order_ids_arr = json_decode( $batch->order_ids, true );
						if ( ! empty( $order_ids_arr ) && ! in_array( $order_id, $order_ids_arr ) ) {
							$order_ids_arr[]   = $order_id;
							$data['order_ids'] = json_encode( $order_ids_arr, JSON_NUMERIC_CHECK );
						}
					} else {
						$data['order_ids'] = json_encode( array( $order_id ) );
					}

					// Set ID for DB
					$where = array(
						'id' => $row['id']
					);

				} elseif ( $operation_type == 'restore' ) {
					// Check how many times the batch is present in the order
					$order                   = wc_get_order( $order_id );
					$same_batch_number_count = array();

					foreach ( $order->get_items() as $item ) {
						$item_batch_numbers = $item->get_meta( '_wcpbn_data' );
						foreach ( $item_batch_numbers as $row ) {
							if ( $row['id'] == $batch->id ) {
								$same_batch_number_count[] = 1;
							}
						}
					}

					// If appears just 1 time we remove the order ID from the 'order_ids' column
					if ( count( $same_batch_number_count ) == 1 || $movement_type == 'all_items' ) {
						// Remove order ID
						if ( ! empty( $batch->order_ids ) ) {
							$order_ids_arr = json_decode( $batch->order_ids, true );
							if ( ! empty( $order_ids_arr ) && in_array( $order_id, $order_ids_arr ) ) {
								foreach ( $order_ids_arr as $key => $value ) {
									if ( $order_id == $value ) {
										unset( $order_ids_arr[$key] );
									}
								}
								sort( $order_ids_arr ); // fix array key gaps
							}
							$data['order_ids'] = json_encode( $order_ids_arr );
						}
					}

					// Set status active if stock available
					if ( $data['quantity_available'] != 0 ) {
						$data['status'] = 'active';
					}

					// Set ID for DB
					$where = array(
						'id' => $batch->id
					);
				}

				// Update database
				$query = wpo_wcpbn_update_batch( $where, $data );
				
				// sync product stock
				if ( $query ) {
					$all_shared_ids = wpo_wcpbn_get_batch_shared_product_ids( $batch->id );
					if ( ! empty( $all_shared_ids ) ) {
						foreach ( $all_shared_ids as $id ) {
							$product = wc_get_product( $id );
							if ( ! empty( $product ) ) {
								wpo_wcpbn_sync_product_stock_with_batches( $product );
							}
						}	
					}
				}
			}
		}

		/**
		 * Data operation types for batch numbers
		 *
		 * @return bool
		 */
		private function data_operation_types( $type ) {
			$allowed       = false;
			$allowed_types = array( 'reduce', 'update', 'restore' );
			if ( in_array( $type, $allowed_types ) ) {
				$allowed = true;
			}
			return $allowed;
		}

		/**
		 * Depending on WooCommerce Stock Management setting set the allowed order status for stock reduce
		 *
		 * @return array
		 */
		public function get_allowed_stock_order_status( $stock_type ) {
			if ( $stock_type === 'reduce' ) {
				$allowed_status = array( 'completed', 'processing', 'on-hold' );
			} elseif ( $stock_type === 'restore' ) {
				$allowed_status = array( 'cancelled', 'pending' );
			}
			return apply_filters( 'wpo_wcpbn_allowed_stock_order_status', $allowed_status, $stock_type );
		}

		public function include_in_formatted_item_meta( $formatted_meta, $item ) {
			$setting = WPO_WCPBN()->classes['settings']->options->get_option( 'display_batch_data_in_formatted_item_meta' );

			if ( ! empty( $item ) ) {
				$order = $item->get_order();

				if ( empty( $order ) ) {
					return $formatted_meta;
				}

				$item_id = ( $order->get_type() == 'shop_order_refund' ) ? $item->get_meta( '_refunded_item_id' ) : $item->get_id();

				// add only to certain order status
				if ( in_array( $order->get_status(), $this->get_allowed_stock_order_status( 'reduce' ) ) ) {
					$item_batch_numbers = wc_get_order_item_meta( $item_id, '_wcpbn_data', true );
					
					if ( empty( $item_batch_numbers ) ) {
						return $formatted_meta;
					}

					foreach ( $item_batch_numbers as $batch_number ) {
						$value = '';

						if ( in_array( $setting, array( 'number', 'number+date' ) ) ) {
							$item_quantity = wpo_wcpbn_format_stock_amount( $batch_number['quantity_sold'] );
							if ( $order->get_type() == 'shop_order_refund' ) {
								$item_quantity = "-{$item_quantity}";
							}
							$value .= $batch_number['batch_number'].' ('.$item_quantity.'x)';
						}
						
						if ( in_array( $setting, array( 'date', 'number+date' ) ) ) {
							$db_row = wpo_wcpbn_get_batches( array( 'id' => $batch_number['id'] ) );
							if ( ! empty( $db_row ) && $db_row[0]->date_expiry != '0000-00-00 00:00:00' && $db_row[0]->date_expiry != '1000-01-01 00:00:00' ) {
								if ( ! empty( $value ) ) {
									$value .= ' - ';
								}
								$value .= get_date_from_gmt( $db_row[0]->date_expiry, wpo_wcpbn_date_format() );
							}
						}

						if ( ! empty( $value ) ) {
							$formatted_meta[] = (object) array(
								'key' 			=> 'batch_number',
								'display_key'	=> __( 'Batch', 'woocommerce-product-batch-numbers' ),
								'value'			=> $value,
								'display_value'	=> '<p>'.$value.'</p>'
							);
						}
					}
				}
			}

			return $formatted_meta;
		}

		private function save_order_items_single_batch_meta( $order_id ) {
			if ( ! empty( $order_id ) ) {
				$order = wc_get_order( $order_id );
				if ( ! empty( $order ) ) {
					foreach ( $order->get_items() as $item_id => $item ) {
						$item_batch_numbers = wc_get_order_item_meta( $item_id, '_wcpbn_data', true );
						if ( ! empty( $item_batch_numbers ) ) {
							// save batch single meta
							$c = 0;
							foreach ( $item_batch_numbers as $batch_number ) {
								$c++;
								// save batch number to single meta (to be used by third party export plugins)
								wc_update_order_item_meta( $item->get_id(), '_wcpbn_batch_'.$c, apply_filters( 'wpo_wcpbn_order_item_single_meta_value', $batch_number['batch_number'], $batch_number ) );
								// hook for third party developers to save additional meta
								do_action( 'wpo_wcpbn_after_save_order_items_single_batch_meta', $batch_number, $item, $order );
							}
						}
					}
				}
			}
		}

		/**
		 * Hides batch item meta from WC Order.
		 *
		 * @return array
		 */
		public function hide_batch_itemmeta_wc_order( $arr ) {
			$arr[] = 'batch_number';
			$arr[] = '_wcpbn_batch_1';
			$arr[] = '_wcpbn_batch_2';
			$arr[] = '_wcpbn_batch_3';
			$arr[] = '_wcpbn_batch_4';
			$arr[] = '_wcpbn_batch_5';

			return apply_filters( 'wpo_wcpbn_hide_wc_order_item_meta', $arr );
		}

		/**
		 * Cleans the batch array to save on order item meta.
		 *
		 * @return array
		 */
		public function clean_order_item_batch_for_meta( $arr ) {
			if ( ! is_array( $arr ) ) {
				return array();
			}

			foreach ( $arr as $id => $batch ) {
				if ( ! is_array( $batch ) ) {
					continue;
				}

				foreach ( $batch as $key => $value ) {
					if ( in_array( $key, array( 'date_created', 'user_id', 'product_id', 'variation_id', 'shared_ids', 'quantity', 'quantity_available', 'status', 'order_ids' ) ) ) {
						unset( $arr[$id][$key] );
					}
				}
			}

			return $arr;
		}

		/**
		 * Adds batch reduce stock support for WooCommerce Subscriptions renewal order creation.
		 *
		 * @return object
		 */
		public function wcs_reduce_stock( $renewal_order, $subscription ) {
			if ( ! empty( $renewal_order ) ) {
				$this->reduce_stock( $renewal_order );
			}

			return $renewal_order;
		}
		
		/**
		 * Reduce stock on RedSys Gateway action.
		 *
		 * @return void
		 */
		public function rgw_reduce_stock( $order, $datos_decodificados, $redsys_gateway ) {
			if ( ! empty( $order ) && empty( get_transient( "wpo_wcpbn_rgw_reduce_stock_order_{$order->get_id()}" ) ) ) {
				set_transient( "wpo_wcpbn_rgw_reduce_stock_order_{$order->get_id()}", 'yes', HOUR_IN_SECONDS );
				$this->reduce_stock( $order );
			}
		}

	}

}