<?php
/**
 * Settings Overview List Table Class
 *
 * @since 1.1.0
 */

namespace WPO\WC\Product_Batch_Numbers;

use WPO\WC\Product_Batch_Numbers\wcpbnBatchNumberStatus;

/**
 * If this file is called directly, abort.
 *
 * @since 1.1.0
 */
if ( ! defined( 'ABSPATH' ) ) {
	die;
}

if ( ! class_exists( 'WP_List_Table' ) ) {
	require_once( ABSPATH . 'wp-admin/includes/class-wp-list-table.php' );
}

if ( ! class_exists( 'wcpbnOverviewListTable' ) ) {

	class wcpbnOverviewListTable extends \WP_List_Table {

		private $counts;
		private $available_status;

		/**
		 * Constructor.
		 *
		 * @return void
		 */
		public function __construct() {
			//Set parent defaults
			parent::__construct( array(
				'singular' => 'batch',   // singular name of the listed records
				'plural'   => 'batches', // plural name of the listed records
				'ajax'     => false      // does this table support ajax?
			) );

			$this->available_status = wcpbnBatchNumberStatus::batch_number_status();
			$this->counts = array();
			foreach ( $this->available_status as $key => $value ) {
				$this->counts[$key] = 0;
			}
			$this->counts['all'] = 0;

			add_filter( 'woocommerce_product_data_store_cpt_get_products_query', [ $this, 'handle_custom_query_vars' ], 10, 2 );

			$this->get_counts();
		}

		/**
		 * Get status results count.
		 *
		 * @return void
		 */
		public function get_counts() {
			$keyword = $this->get_search();
			
			foreach ( $this->counts as $key => $value ) {
				if ( $key != 'all' ) {
					$search_key = $search_by = isset( $_GET['search_by'] ) ? esc_attr( $_GET['search_by'] ) : '';
					
					$args = [
						'status' => $key,
						'search' => [
							'term' => $keyword,
							'type' => 'pattern'
						],
					];

					$batch_ids = [];
					
					switch ( $search_by ) {
						case 'batch':
							$search_key = 'batch_number';
							break;
						case 'product_title':
							if ( ! empty( $keyword ) ) {
								$batch_ids = $this->get_batch_numbers_by_product_title( $keyword );
							}
							$search_key = 'id';
							break;
						case 'product_meta':
							if ( ! empty( $keyword ) ) {
								$batch_ids = $this->get_batch_numbers_by_product_meta( $keyword );
							}
							$search_key = 'id';
							break;
						case 'product_id':
							if ( ! empty( $keyword ) ) {
								$batch_ids = wpo_wcpbn_get_product_batch_ids( [ $keyword ] );
							}
							$search_key = 'id';
							break;
						case 'product_sku':
							if ( ! empty( $keyword ) ) {
								$product_ids = $this->get_product_ids_by_partial_sku( sanitize_text_field( $keyword ) );
								$batch_ids   = wpo_wcpbn_get_product_batch_ids( $product_ids );
							}
							$search_key = 'id';
							break;
						case 'id':
							$args['search']['type'] = 'exact';
							break;
					}
					
					
					
					if ( ! empty( $batch_ids ) ) {
						$args['search']['term'] = $batch_ids;
						$args['search']['type'] = 'in';
					}

					$args['search']['keys'] = [ $search_key ];
										
					$result = wpo_wcpbn_get_batches( apply_filters( 'wpo_wcpbn_overview_list_get_counts_args', $args ) );
					
					$this->counts[$key] = count( $result );
				}
			}
			foreach ( $this->counts as $key => $value ) {
				if ( $key != 'all' ) {
					$this->counts['all'] += (int) $value;
				}
			}
		}

		/**
		 * Get table views.
		 *
		 * @return array
		 */
		public function get_views() {
			$base    = $this->get_base_url();
			$current = isset( $_GET['status'] ) ? sanitize_key( $_GET['status'] ) : '';
			$is_all  = empty( $current ) || ( 'all' === $current );
			$counts  = array();

			foreach ( $this->counts as $key => $value ) {
				$counts[$key] = '&nbsp;<span class="count">(' . esc_html( $this->counts[$key] ) . ')</span>';
			}

			$count_links = array();
			foreach ( $counts as $key => $value ) {
				if ( $key == 'all' ) {
					$count_links[$key] = sprintf( '<a href="%s"%s>%s</a>', esc_url( remove_query_arg( 'status', $base ) ), $is_all ? ' class="current"' : '', __( 'All', 'woocommerce-product-batch-numbers' ) . $value );
				} else {
					$count_links[$key] = sprintf( '<a href="%s"%s>%s</a>', esc_url( add_query_arg( 'status', $key, $base ) ), $key === $current ? ' class="current"' : '', $this->available_status[$key] . $value );
				}
			}
	
			return $count_links;
		}

		/**
		 * Defines the search box.
		 *
		 * @return void
		 */
		public function search_box( $text, $input_id ) {
			$input_id = $input_id . '-search-input';
			if ( ! empty( $_REQUEST['orderby'] ) ) {
				echo '<input type="hidden" name="orderby" value="' . esc_attr( $_REQUEST['orderby'] ) . '" />';
			}
			if ( ! empty( $_REQUEST['order'] ) ) {
				echo '<input type="hidden" name="order" value="' . esc_attr( $_REQUEST['order'] ) . '" />';
			}
			if ( ! empty( $_REQUEST['tab'] ) ) {
				echo '<input type="hidden" name="tab" value="' . esc_attr( $_REQUEST['tab'] ) . '" />';
			}
			?>
				<p class="search-box">
					<label class="screen-reader-text" for="<?php echo $input_id; ?>"><?php echo $text; ?>:</label>
					<input type="search" id="<?php echo $input_id; ?>" name="s" value="<?php _admin_search_query(); ?>" required />
					<?php submit_button( $text, 'button', false, false, array( 'ID' => 'search-submit' ) ); ?>
				</p>
			<?php
		}

		public function column_cb( $item ) {
			return sprintf( '<input type="checkbox" name="batch_id[]" value="%s" />', $item->id );
		}

		/**
		 * Defines the column batch number.
		 *
		 * @return void
		 */
		public function column_id( $item ) {
			echo $item->id;
		}

		/**
		 * Defines the column batch number.
		 *
		 * @return void
		 */
		public function column_batch_number( $item ) {
			echo '<strong>'.$item->batch_number.'</strong>';
		}

		/**
		 * Defines the column supplier.
		 *
		 * @return void
		 */
		public function column_supplier( $item ) {
			echo $item->supplier;
		}
		
		/**
		 * Defines the column po_number.
		 *
		 * @return void
		 */
		public function column_po_number( $item ) {
			$po = wpo_wcpbn_get_batch_purchase_order( absint( $item->id ) );
			if ( $po ) {
				echo $po->po_number;
			}
		}

		/**
		 * Defines the column product.
		 *
		 * @return void
		 */
		public function column_product( $item ) {
			$shared_ids = wpo_wcpbn_get_batch_shared_product_ids( $item->id );
			
			if ( empty( $shared_ids ) ) {
				return;
			}
			
			foreach ( $shared_ids as $product_id ) {
				$product = wc_get_product( $product_id );
				if ( ! empty( $product ) ) {
					$pid = is_callable( [ $product, 'get_parent_id' ] ) && $product->get_parent_id() != 0 ? $product->get_parent_id() : $product_id;
					printf(
						'<a href="%1$s">%2$s (#%3$s)</a><br>',
						get_edit_post_link( wpo_wcpbn_wpml_object_id( $pid ) ),
						$product->get_name(),
						$product_id
					);
				}
			}
		}

		/**
		 * Defines the column sku.
		 *
		 * @return void
		 */
		public function column_sku( $item ) {
			$all_shared_ids = wpo_wcpbn_get_batch_shared_product_ids( $item->id );
			
			foreach ( $all_shared_ids as $id ) {
				$product = wc_get_product( $id );
				if ( ! empty( $product ) && ! empty( $sku = $product->get_sku() ) ) {
					echo $sku.' (#'.$id.')<br>';
				}
			}
		}

		/**
		 * Defines the column location.
		 *
		 * @return void
		 */
		public function column_location( $item ) {
			if ( property_exists( $item, 'location' ) ) {
				echo $item->location;
			}
		}

		/**
		 * Defines the column quantity.
		 *
		 * @return void
		 */
		public function column_quantity( $item ) {
			echo wpo_wcpbn_format_stock_amount( $item->quantity );
		}

		/**
		 * Defines the column quantity available.
		 *
		 * @param object|array $item
		 *
		 * @return void
		 */
		public function column_quantity_available( $item ): void {
			$quantity           = wpo_wcpbn_force_stock_amount_type( $item->quantity );
			$quantity_available = wpo_wcpbn_force_stock_amount_type( $item->quantity_available );

			if ( $quantity_available <= $this->percentage_calculation( 10, $quantity ) && $quantity_available >= $this->percentage_calculation( 5, $quantity ) ) {
				$color = '#dba617';
			} elseif ( $quantity_available <= $this->percentage_calculation( 5, $quantity ) ) {
				$color = '#f86368';
			} else {
				$color = '#00ba37';
			}

			printf( '<input type="number" name="batches[%1$s][quantity_available]" min="0" placeholder="0" style="color:%2$s;" value="%3$s" data-value="%3$s">', $item->id, $color, wpo_wcpbn_format_stock_amount( $quantity_available ) );
		}

		/**
		 * Defines the column date expiry.
		 *
		 * @param object|array $item
		 *
		 * @return void
		 */
		public function column_date_expiry( $item ): void {
			if ( '0000-00-00 00:00:00' !== $item->date_expiry && '1000-01-01 00:00:00' !== $item->date_expiry ) {
				$batch_date = strtotime( $item->date_expiry );
				$class      = wpo_wcpbn_get_batch_expiry_date_class( $item );
			}

			$batch_date = ! empty( $batch_date ) ? date_i18n( 'Y-m-d', $batch_date ) : '';

			printf( '<input type="date" name="batches[%1$s][date_expiry]" class="%2$s" value="%3$s" data-value="%3$s" />', $item->id, $class ?? '', $batch_date );
		}

		/**
		 * Defines the column status.
		 *
		 * @return void
		 */
		public function column_status( $item ) {
			foreach ( $this->available_status as $slug => $name ) {
				if ( $item->status == 'active' && $item->status == $slug ) {
					$background_color = '#c7e8ca';
					$text             = $name;
				} elseif ( $item->status == 'archived' && $item->status == $slug ) {
					$background_color = '#ccc';
					$text             = $name;
				}
			}
			printf( '<span style="background-color:%1$s;">%2$s</span>', $background_color, $text );
		}

		/**
		 * Defines the column actions.
		 *
		 * @return void
		 */
		public function column_actions( $item ) {
			echo '<a href="'.admin_url( 'edit.php?post_type=product&page=wpo-batch-numbers&tab=tools&batch='.$item->id ).'" class="batch-tools"><span class="dashicons dashicons-admin-tools"></span></a>';
			echo '<a href="#" class="delete-batch" data-id="'.$item->id.'"><span class="dashicons dashicons-trash"></span></a>';
		}

		/**
		 * Method to add custom columns
		 *
		 * @return void
		 */
		public function column_default( $item, $column_name ) {
			do_action( 'wpo_wcpbn_overview_list_custom_column', $column_name, $item );
		}

		/**
		 * Defines all available columns.
		 *
		 * @return string
		 */
		public function get_columns() {
			$columns = array(
				'cb'                 => '<input type="checkbox" />', // render a checkbox instead of text
				'id'                 => __( 'ID', 'woocommerce-product-batch-numbers' ),
				'batch_number'       => __( 'Batch', 'woocommerce-product-batch-numbers' ),
				'supplier'           => __( 'Supplier', 'woocommerce-product-batch-numbers' ),
				'po_number'          => __( 'PO number', 'woocommerce-product-batch-numbers' ),
				'product'            => __( 'Product(s)', 'woocommerce-product-batch-numbers' ),
				'sku'                => __( 'SKU', 'woocommerce-product-batch-numbers' ),
				'quantity'           => __( 'Initial Quantity', 'woocommerce-product-batch-numbers' ),
				'quantity_available' => __( 'Quantity Available', 'woocommerce-product-batch-numbers' ),
				'date_expiry'        => WPO_WCPBN()->classes['settings']->options->get_batch_date_label( WPO_WCPBN()->classes['settings']->options->get_option( 'batch_date_label' ) ),
				'status'             => __( 'Status', 'woocommerce-product-batch-numbers' ),
				'actions'            => __( 'Actions', 'woocommerce-product-batch-numbers' ),
			);

			if ( wc_string_to_bool( WPO_WCPBN()->classes['settings']->options->get_option( 'stock_locations' ) ) ) {
				$columns = array_slice( $columns, 0, count( $columns ) - 5, true ) + array( 'location' => __( 'Location', 'woocommerce-product-batch-numbers' ) ) + array_slice( $columns, 0, count( $columns ), true );
			}

			return apply_filters( 'wpo_wcpbn_overview_list_columns', $columns );
		}

		/**
		 * Defines the sortable columns.
		 *
		 * @return array
		 */
		public function get_sortable_columns() {
			$sortable = array(
				'id'                 => array( 'id', true ),
				'batch_number'       => array( 'batch_number', true ),
				'supplier'           => array( 'supplier', true ),
				'quantity_available' => array( 'quantity_available', true ),
				'date_expiry'        => array( 'date_expiry', true ),
				'order_ids'          => array( 'order_ids', true ),
			);

			if ( wc_string_to_bool( WPO_WCPBN()->classes['settings']->options->get_option( 'stock_locations' ) ) ) {
				$sortable['location'] = array( 'location', true );
			}

			return $sortable;
		}

		/**
		 * Gets search string from request.
		 *
		 * @return string
		 */
		public function get_search() {
			return ! empty( $_GET['s'] ) ? urldecode( trim( $_GET['s'] ) ) : false;
		}
		
		/**
		 * Gets search by from request.
		 *
		 * @return string
		 */
		public function get_search_by() {
			return ! empty( $_GET['search_by'] ) ? esc_attr( $_GET['search_by'] ) : false;
		}

		/**
		 * Gets the bulk actions.
		 *
		 * @return array
		 */
		function get_bulk_actions(): array {
			return array(
				'delete' => __( 'Delete', 'woocommerce-product-batch-numbers' ),
				'update' => __( 'Update', 'woocommerce-product-batch-numbers' ),
			);
		}

		/**
		 * Processes the bulk actions.
		 *
		 * @return void
		 */
		function process_bulk_action(): void {
			if ( empty( $_REQUEST['_wpnonce'] ) || empty( $_REQUEST['batch_id'] ) ) {
				return;
			}

			$data   = stripslashes_deep( $_REQUEST );
			$nonce  = $data['_wpnonce'];
			$action = 'bulk-' . $this->_args['plural'];

			if ( ! wp_verify_nonce( $nonce, $action ) ) {
				wp_die( __( "You don't have permission to delete this batch (nonce expired or invalid)", 'woocommerce-product-batch-numbers' ) );
			}

			switch ( $this->current_action() ) {
				case 'delete':
					foreach ( $data['batch_id'] as $batch_id ) {
						wpo_wcpbn_delete_batch( absint( $batch_id ), $data );
					}
					break;
				case 'update':
					$this->update_batches( $data );
					break;
			}
		}

		private function update_batches( array $data ): void {
			if ( empty( $data['batches'] ) ) {
				return;
			}

			$update_data = array();
			foreach ( $data['batches'] as $batch_id => $batch ) {
				// Pass unselected batches.
				if ( ! in_array( $batch_id, $data['batch_id'] ) ) {
					continue;
				}

				foreach ( $batch as $column => $value ) {
					switch ( $column ) {
						case 'quantity_available':
							$update_data[ $batch_id ][ $column ] = is_numeric( $value ) && 0 < absint( $value ) ? wpo_wcpbn_force_stock_amount_type( absint( $value ) ) : 0;
							break;
						case 'date_expiry':
							if ( empty( $value ) ) {
								$update_data[ $batch_id ][ $column ] = $value;
								break;
							}

							$date_object = \DateTime::createFromFormat( 'Y-m-d', $value );

							if ( $date_object ) {
								$update_data[ $batch_id ][ $column ] = $value;

								$current_date                       = new \DateTime( current_time( 'Y-m-d' ) );
								$update_data[ $batch_id ]['status'] = $date_object->format( 'Y-m-d' ) >= $current_date->format( 'Y-m-d' ) ? 'active' : 'archived';
							}
							break;
					}
				}
			}

			if ( ! empty( $update_data ) ) {
				wpo_wcpbn_update_batches( $update_data );
			}
		}

		/**
		 * Prepare items for table.
		 * 
		 * @uses $this->_column_headers
		 * @uses $this->items
		 * @uses $this->get_columns()
		 * @uses $this->get_sortable_columns()
		 * @uses $this->get_pagenum()
		 * @uses $this->set_pagination_args()
		 */
		public function prepare_items() {
			$user          = get_current_user_id();
			$screen        = get_current_screen();
			$screen_option = $screen->get_option( 'per_page', 'option' );
			$columns       = $this->get_columns();
			$hidden        = array();
			$sortable      = $this->get_sortable_columns();
			
			$this->_column_headers = array( $columns, $hidden, $sortable );

			$this->process_bulk_action();
			
			$defaults                 = [];
			$defaults['request']      = stripslashes_deep( $_GET );
			$defaults['search']       = $this->get_search();
			$defaults['search_by']    = $this->get_search_by();
			$defaults['current_page'] = $this->get_pagenum();
			$defaults['status']       = isset( $defaults['request']['status'] ) ? sanitize_text_field( $defaults['request']['status'] ) : [ 'active', 'archived' ];

			if ( is_string( $screen_option ) ) {
				$defaults['per_page'] = get_user_meta( $user, $screen_option, true );
				if ( empty( $defaults['per_page'] ) || intval( $defaults['per_page'] ) < 1 ) {
					$defaults['per_page'] = $screen->get_option( 'per_page', 'default' );
				}
			} else {
				$defaults['per_page'] = 10;
			}

			if ( ! empty( $defaults['search'] ) && ! empty( $defaults['search_by'] ) ) {
				switch ( $defaults['search_by'] ) {
					default:
					case 'batch':
						$args = array(
							'search' => array( 'keys' => [ 'batch_number' ], 'term' => sanitize_text_field( $defaults['search'] ), 'type' => 'pattern' ),
							'status' => $defaults['status'],
						);
						$data = wpo_wcpbn_get_batches( $args );
						break;
					case 'supplier':
						$args = array(
							'search' => array( 'keys' => [ 'supplier' ], 'term' => sanitize_text_field( $defaults['search'] ), 'type' => 'pattern' ),
							'status' => $defaults['status'],
						);
						$data = wpo_wcpbn_get_batches( $args );
						break;
					case 'id':
						if (  0 < intval( $defaults['search'] )  ) {
							$data = wpo_wcpbn_get_batches( array( 'id' => intval( $defaults['search'] ) ) );
						} else {
							$data = [];
						}
						break;
					case 'location':
						$data = wpo_wcpbn_get_batches( array( 'location' => sanitize_text_field( $defaults['search'] ) ) );
						break;
					case 'product_sku':
						$product_ids = $this->get_product_ids_by_partial_sku( sanitize_text_field( $defaults['search'] ) );

						if ( is_array( $product_ids ) && ! empty( $product_ids ) ) {
							$args = array(
								'shared_ids' => $product_ids,
								'status'     => $defaults['status'],
							);
							$data = wpo_wcpbn_get_batches( $args );
						} else {
							$data = [];
						}
						break;
					case 'product_id':
						if ( is_numeric( $defaults['search'] ) ) {
							$args = array(
								'shared_ids' => [ intval( $defaults['search'] ) ],
								'status'     => $defaults['status'],
							);
							$data = wpo_wcpbn_get_batches( $args );
						} else {
							$data = [];
						}
						break;
					case 'product_title': 
						$product_ids = $this->get_product_ids_by_title( sanitize_text_field( $defaults['search'] ) );
						if ( ! empty( $product_ids ) ) {
							$args = [
								'shared_ids' => $product_ids,
								'status'     => $defaults['status'],
							];
							$data = wpo_wcpbn_get_batches( $args );
						} else {
							$data = [];
						}
						break;
					case 'product_meta': 
						$product_ids = $this->get_product_ids_by_meta( sanitize_text_field( $defaults['search'] ) );
						if ( ! empty( $product_ids ) ) {
							$args = [
								'shared_ids' => $product_ids,
								'status'     => $defaults['status'],
							];
							$data = wpo_wcpbn_get_batches( $args );
						} else {
							$data = [];
						}
						break;
				}
			} else {
				$args = array(
					'status'   => $defaults['status'],
					'order'    => isset( $defaults['request']['order'] ) ? strtoupper( sanitize_text_field( $defaults['request']['order'] ) ) : apply_filters( 'wpo_wcpbn_overview_list_order', 'ASC' ),
					'order_by' => isset( $defaults['request']['orderby'] ) ? sanitize_text_field( $defaults['request']['orderby'] ) : apply_filters( 'wpo_wcpbn_overview_list_orderby', 'date_expiry' ),
					'limit'    => $defaults['per_page'],
					'offset'   => $defaults['per_page'] * ( $defaults['current_page'] - 1 ),
				);
				$data = wpo_wcpbn_get_batches( $args );
			}
			
			$total_items             = wpo_wcpbn_get_batches( [ 'status' => $defaults['status'] ] );
			$defaults['total_items'] = count( $total_items );
			
			$this->items = apply_filters( 'wpo_wcpbn_overview_list_items_data', $data, $defaults );
			
			$this->set_pagination_args( array(
				'total_items' => $defaults['total_items'],                                // WE have to calculate the total number of items
				'per_page'    => $defaults['per_page'],                                   // WE have to determine how many items to show on a page
				'total_pages' => ceil( $defaults['total_items'] / $defaults['per_page'] ) // WE have to calculate the total number of pages
			) );
		}

		/**
		 * Message when no items in table.
		 *
		 * @return void
		 */
		public function no_items() {
			_e( 'No batch numbers have been found.', 'woocommerce-product-batch-numbers' );
		}

		/**
		 * Calculates the percentage of a given number.
		 *
		 * @return int
		 */
		private function percentage_calculation( $percentage, $total ) {
			return ( $percentage / 100 ) * $total;
		}

		/**
		 * Return product IDs based on product title.
		 *
		 * @param string $product_title Product title.
		 * @return array
		 */
		public function get_product_ids_by_title( $product_title ) {
			// Get products.
			$products = wc_get_products( array(
				'status' 	 => 'publish',
				'return' 	 => 'ids',
				'like_title' => $product_title,
				'limit' 	 => -1
			) );
			
			// Get product variations.	
			$variation_products = wc_get_products( array(
				'status' 	 => 'publish',
				'return' 	 => 'ids',
				'like_title' => $product_title,
				'type' 		 => 'variation',
				'limit' 	 => -1
			) );
					
			return array_merge( $products, $variation_products );	
		}
		
		/**
		 * Return product IDs based on product meta.
		 *
		 * @param string $product_meta Product meta.
		 * @return array
		 */
		public function get_product_ids_by_meta( $product_meta ) {
			// Get products.
			$products = wc_get_products( array(
				'status' 	=> 'publish',
				'return' 	=> 'ids',
				'like_meta' => $product_meta,
				'limit' 	=> -1
			) );
			
			// Get product variations.	
			$variation_products = wc_get_products( array(
				'status' 	=> 'publish',
				'return' 	=> 'ids',
				'like_meta' => $product_meta,
				'type' 		=> 'variation',
				'limit' 	=> -1
			) );
					
			return array_merge( $products, $variation_products );	
		}

		/**
		 * Return batch IDs based on product title.
		 *
		 * @param string $product_title Product title.
		 * @return array
		 */
		public function get_batch_numbers_by_product_title( $product_title ) {
			$product_ids = $this->get_product_ids_by_title( $product_title );
			return wpo_wcpbn_get_product_batch_ids( $product_ids );
		}
		
		/**
		 * Return batch IDs based on product meta.
		 *
		 * @param string $product_meta Product meta.
		 * @return array
		 */
		public function get_batch_numbers_by_product_meta( $product_meta ) {
			$product_ids = $this->get_product_ids_by_meta( $product_meta );
			return wpo_wcpbn_get_product_batch_ids( $product_ids );
		}

		/**
		 * Modify query results to search.
		 * 
		 */
		function handle_custom_query_vars( $query, $args ) {
			// by title
			if ( isset( $args['like_title'] ) && ! empty( $args['like_title'] ) ) {
				$query['s']              = esc_attr( $args['like_title'] );
				$query['search_columns'] = 'post_title';
				
			// by meta
			} elseif ( isset( $args['like_meta'] ) && ! empty( $args['like_meta'] ) ) {
				$meta = explode( '::', esc_attr( $args['like_meta'] ) );
				if ( ! empty( $meta ) && is_array( $meta ) ) {
					$meta_key   = esc_attr( $meta[0] );
					$meta_value = esc_attr( $meta[1] );
					
					$keys   = [];
					$keys[] = $meta_key;
					
					if ( $meta_key[0] == '_' ) {
						$keys[] = ltrim( $meta_key, '_' );
					} else {
						$keys[] = "_{$meta_key}";
					}
					
					$query['meta_query']['relation'] = 'OR';
					
					foreach ( $keys as $key ) {
						$query['meta_query'][] = array(
							'key'   => $key,
							'value' => $meta_value,
						);	
					}
				}
			}
			return $query;
		}
		
		/**
		 * Returns all products that match or partially match the code 'SKU code' provided
		 * 
		 * @param	string $sku	The code that will be searched
		 * 
		 * @return array of product IDs that match the 'SKU code' provided
		 */
		function get_product_ids_by_partial_sku( $sku ) {
			$args = array(
				'sku'    => $sku,
				'return' => 'ids',
				'limit'  => -1,
			);

			$products = wc_get_products( $args );

			return $products;
		}
	}
}

return new wcpbnOverviewListTable();