<?php
/**
 * Product Listing Class
 *
 * @since 1.2.0
 */

namespace WPO\WC\Product_Batch_Numbers;

/**
 * If this file is called directly, abort.
 *
 * @since 1.0.0
 */
if ( ! defined( 'ABSPATH' ) ) {
	die;
}

if ( ! class_exists( 'wcpbnProductListing' ) ) {

	class wcpbnProductListing {

		protected        $orderby;
		protected static $_instance = null;

		/**
		 * Instance
		 */
		public static function instance() {
			if ( is_null( self::$_instance ) ) {
				self::$_instance = new self();
			}
			return self::$_instance;
		}

		/**
		 * Construct
		 */
		public function __construct() {
			// show/hide column based on option value
			if ( ! empty( WPO_WCPBN()->classes['settings']->options->get_option( 'display_batch_column_products' ) ) && WPO_WCPBN()->classes['settings']->options->get_option( 'display_batch_column_products' ) != 'no' ) {
				add_filter( 'manage_edit-product_columns', array( $this, 'batch_listing_column' ), 10, 1 );
				add_action( 'manage_posts_custom_column', array( $this, 'populate_batch_column' ), 10, 2 );
			}

			// get batch order by priority
			$this->orderby = ( $batch_stock_reduction_priority = WPO_WCPBN()->classes['settings']->options->get_option( 'batch_prioritization' ) ) ? $batch_stock_reduction_priority : 'id';

			add_action( 'woocommerce_product_quick_edit_save', array( $this, 'save_column_data' ), 10, 1 );
			add_action( 'wp_ajax_get_product_batches', array( $this, 'get_product_batches' ) );
		}

		/**
		 * Remove column from post type 'product' listing.
		 *
		 * @return array
		 */
		public function batch_listing_column( $columns ) {			
			$offset          = 6;
			$updated_columns = array_slice( $columns, 0, $offset, true) +
			array( 'product_batch_numbers' => esc_html__( 'Batches', 'woocommerce-product-batch-numbers' ) ) +
			array_slice( $columns, $offset, null, true );
			return $updated_columns;
		}

		/**
		 * Populate 'Stock at locations' column.
		 *
		 * @return void
		 */
		public function populate_batch_column( $column_name, $post_or_product_object ) {
			if ( $column_name != 'product_batch_numbers' ) {
				return;
			}
			
			$product = ( $post_or_product_object instanceof \WP_Post ) ? wc_get_product( $post_or_product_object->ID ) : $post_or_product_object;
			if ( ! is_object( $product ) && is_numeric( $product ) ) {
				$product = wc_get_product( absint( $product ) );
			}

			// main product
			$this->column_batches_list( $product->get_id() );

			// product variations
			$variation_ids = $product->get_children();

			foreach ( $variation_ids as $variation_id ) {
				$this->column_batches_list( $variation_id );
			}
		}
		
		/**
		 * Creates the batches lists for the column.
		 *
		 * @return void
		 */
		private function column_batches_list( $product_id ) {
			if ( is_object( $product_id ) ) {
				return;
			}

			$product = wc_get_product( $product_id );

			if ( empty( $product ) ) {
				return;
			}

			$product_type = $product->get_type();

			if ( ! in_array( $product_type, array( 'simple', 'variation', 'bundle', 'subscription', 'variable-subscription' ) ) ) {
				return;
			}

			$data_format = WPO_WCPBN()->classes['settings']->options->get_option( 'display_batch_column_products' );
			wpo_wcpbn_print_product_batch_data_formatted( $product_id, $data_format, true, false );
		}

		/**
		 * Get the product batches for AJAX.
		 *
		 * @return void
		 */
		public function get_product_batches() {
			if ( $_POST && isset( $_POST['action'] ) && isset( $_POST['post_id'] ) && $_POST['action'] == 'get_product_batches' ) {
				$product_id = sanitize_text_field( $_POST['post_id'] );
				$product    = wc_get_product( $product_id );
				
				if ( empty( $product ) ) {
					return;
				}

				$product_batches = wpo_wcpbn_get_batches( [ 'shared_ids' => [ intval( $product_id ) ] ] );

				wp_send_json_success( compact( 'product_batches' ) );
			}
		}

		/**
		 * Saves batch data on the quick edit form.
		 *
		 * @return void
		 */
		public function save_column_data( $product ) {
			if ( isset( $_POST ) && isset( $_POST['wpo-wcpbn-table-expirydate'] ) ) {
				$product_id   = $product->get_id();
				$product_type = $product->get_type();

				if ( ! in_array( $product_type, array( 'simple', 'variable', 'bundle', 'subscription', 'variable-subscription' ) ) ) {
					return;
				}

				if ( $product_batches = wpo_wcpbn_get_batches( [ 'shared_ids' => [ intval( $product_id ) ] ] ) ) {
					foreach ( $product_batches as $batch ) {
						foreach ( $_POST['wpo-wcpbn-table-expirydate'] as $batch_id => $date_expiry ) {
							if ( $batch->id == $batch_id ) {
								$input_date          = sanitize_text_field( $date_expiry );
								$current_date_expiry = date( 'Y-m-d', strtotime( $batch->date_expiry ) );
								if ( $input_date != $current_date_expiry ) {
									wpo_wcpbn_update_batch( $batch->id, [ 'date_expiry' => $input_date ] );
								}
							}
						}
					}
				}

			}
		}

	}

}
