<?php
namespace WPO\WC\Product_Batch_Numbers;

if ( ! defined( 'ABSPATH' ) ) {
	die;
}

if ( ! class_exists( 'wcpbnPurchaseOrders' ) ) {

	class wcpbnPurchaseOrders {

		protected static $_instance = null;

		/**
		 * Instance
		 */
		public static function instance() {
			if ( is_null( self::$_instance ) ) {
				self::$_instance = new self();
			}
			return self::$_instance;
		}

		/**
		 * Construct
		 */
		public function __construct() {
			if ( wc_string_to_bool( WPO_WCPBN()->classes['settings']->options->get_option( 'use_po_prices' ) ) ) {
				add_filter( 'woocommerce_product_get_price', array( $this, 'use_po_price' ), 999, 2 );
				
				if ( wc_string_to_bool( WPO_WCPBN()->classes['settings']->options->get_option( 'limit_cart_qty_to_keep_po_price' ) ) ) {
					add_filter( 'woocommerce_add_to_cart_quantity', array( $this, 'limit_add_to_cart_qty' ), 10, 2 );
					add_filter( 'woocommerce_update_cart_validation', array( $this, 'limit_cart_update_qty' ), 10, 4 );
				}
			}
		}
		
		public function calculate_prices( array $args = array() ): array {
			$calculation_needs = array(
				'regular' => array(
					'cost',
					'price_margin',
				),
				'sale' => array(
					'cost',
					'price_margin',
					'price_discount',
				),
			);
			
			foreach ( array( 'regular', 'sale' ) as $price_type ) {
				$can_calculate_price = wc_string_to_bool( WPO_WCPBN()->classes['settings']->options->get_option( "calculate_{$price_type}_price" ) );
			
				if ( ! $can_calculate_price ) {
					continue;
				}
				
				switch ( $price_type ) {
					case 'regular':
						if ( empty( $args['price_margin'] ) ) {
							continue 2;
						}
						break;
					case 'sale':
						if ( empty( $args['price_discount'] ) ) {
							continue 2;
						}
						break;
				}
				
				$price = 0;
				
				foreach ( $calculation_needs[ $price_type ] as $need ) {
					if ( ! isset( $args[ $need ] ) || 0 == $args[ $need ] ) {
						continue;
					}
					
					if ( 'cost' !== $need && 0 == $price ) {
						continue;
					}
					
					switch ( $need ) {
						case 'cost':
							$price += floatval( $args[ $need ] );
							break;
						case 'price_margin':
							$price *= 1 + ( floatval( $args[ $need ] ) / 100 );
							break;
						case 'price_discount':
							$price *= ( 100 - floatval( $args[ $need ] ) ) / 100;
							break;
					}
				}
				
				if ( 0 != $price ) {
					$args[ "{$price_type}_price" ] = wc_format_decimal( floatval( $price ), '' );
				}
			}
			
			return $args;
		}
		
		public function use_po_price( string $value, object $product ): string {
			if ( empty( $product ) ) {
				return $value;
			}
			
			if ( ! is_callable( array( $product, 'get_type' ) ) ) {
				return $value;
			}
			
			if ( ! in_array( $product->get_type(), array( 'simple', 'variation', 'bundle', 'subscription', 'variable-subscription' ) ) ) {
				return $value;
			}
			
			$priority = ! empty( WPO_WCPBN()->classes['settings']->options->get_option( 'po_price_prioritization' ) ) ? WPO_WCPBN()->classes['settings']->options->get_option( 'po_price_prioritization' ) : 'next_batch';
			$args     = apply_filters( 'wpo_wcpbn_use_po_price_get_product_batches_args', array(
				'shared_ids' => array( $product->get_id() ),
				'status'     => 'active',
				'order_by'   => WPO_WCPBN()->classes['settings']->options->get_option( 'batch_prioritization' ) ? WPO_WCPBN()->classes['settings']->options->get_option( 'batch_prioritization' ) : 'id',
				'order'      => 'ASC',
				'limit'      => 'next_batch' === $priority ? 1 : 100,
			), $product, $priority );
			
			$batches = wpo_wcpbn_get_batches( $args );
			
			if ( empty( $batches ) ) {
				return $value;
			}
			
			$po = null;
			
			switch ( $priority ) {
				case 'next_batch':
				default:
					$batch = reset( $batches );
				
					if ( empty( $batch ) ) {
						return $value;
					}
					
					$po = wpo_wcpbn_get_batch_purchase_order( absint( $batch->id ) );
					break;
				case 'highest_regular_price':
				case 'lowest_regular_price':
					$batch_ids = array_map( function( $value ) {
						return $value->id;
					}, $batches );
					
					$pos = wpo_wcpbn_get_pos_from_batch_ids( $batch_ids );
					
					if ( empty( $pos ) ) {
						break;
					}
					
					usort( $pos, function( $a, $b ) use ( $priority ) {
						if ( false !== strpos( $priority, 'highest' ) ) {
							return $b->regular_price <=> $a->regular_price;
						} else {
							return $a->regular_price <=> $b->regular_price;
						}
					} );
					
					$po = reset( $pos );
					break;
			}
			
			if ( empty( $po ) ) {
				return $value;
			}
			
			$type = 'regular_price';
			if ( 0 != $po->sale_price && floatval( $po->regular_price ) > floatval( $po->sale_price ) ) {
				$type = 'sale_price';
			}
			
			if ( empty( $po->$type ) || 0 == $po->$type ) {
				return $value;
			}
			
			return $po->$type;
		}
		
		public function limit_add_to_cart_qty( $quantity, $product_id ) {
			$product = wc_get_product( $product_id );
			
			$this->limit_qty_to_keep_po_price( $product, $quantity, 'product' );
			
			return $quantity;
		}
		
		public function limit_cart_update_qty( $bool, $cart_item_key, $values, $quantity ) {
			$product = wc_get_product( $values['variation_id'] ? $values['variation_id'] : $values['product_id'] );
			
			if ( $this->limit_qty_to_keep_po_price( $product, $quantity, 'cart' ) ) {
				$bool = false;
			}
			
			return $bool;
		}
		
		/**
		 * Limit the quantity in the Product & Cart pages to keep the PO price
		 *
		 * @param  object  $product
		 * @param  int     $quantity
		 * @param  string  $page      can be 'product' or 'cart'
		 * @return boolean
		 */
		public function limit_qty_to_keep_po_price( object $product, int $quantity, string $page = 'product' ): bool {
			if ( $product && $product->managing_stock() && 0 < absint( $quantity ) ) {
				$message   = null;
				$qty_limit = null;
				
				$args = apply_filters( 'wpo_wcpbn_limit_qty_to_keep_po_price_args', array(
					'shared_ids' => array( $product->get_id() ),
					'status'     => 'active',
					'order_by'   => WPO_WCPBN()->classes['settings']->options->get_option( 'batch_prioritization' ) ? WPO_WCPBN()->classes['settings']->options->get_option( 'batch_prioritization' ) : 'id',
					'order'      => 'ASC',
					'limit'      => 1,
				), $product, $quantity );
				
				$batches = wpo_wcpbn_get_batches( $args );
				
				if ( empty( $batches ) ) {
					return false;
				}
				
				$next_batch = reset( $batches );
				
				if ( empty( $next_batch ) ) {
					return false;
				}
				
				if ( 0 != $next_batch->quantity_available ) {
					$qty_limit = absint( $next_batch->quantity_available );
				}
				
				// we have a quantity limit, let's do other checks
				if ( ! empty( $qty_limit ) && $qty_limit > 0 ) {

					if ( $qty_limit < absint( $quantity ) ) {
						/* translators: 1: product name 2: quantity in stock */
						$message = sprintf( __( 'You cannot add that amount of &quot;%1$s&quot; to the cart because there is not enough stock (%2$s remaining).', 'woocommerce-product-batch-numbers' ), $product->get_name(), wpo_wcpbn_format_stock_amount( $qty_limit ) );
					
					} else {
						$products_qty_in_cart = WC()->cart->get_cart_item_quantities();
						
						if ( isset( $products_qty_in_cart[ $product->get_stock_managed_by_id() ] ) ) {
							$new_qty = ( 'cart' === $page ) ? absint( $quantity ) : absint( $products_qty_in_cart[ $product->get_stock_managed_by_id() ] ) + absint( $quantity );
							
							if ( $qty_limit < $new_qty ) {
								$stock_quantity_in_cart = $products_qty_in_cart[ $product->get_stock_managed_by_id() ];
								$wp_button_class        = wc_wp_theme_get_element_class_name( 'button' ) ? ' ' . wc_wp_theme_get_element_class_name( 'button' ) : '';
								$message                = sprintf(
									'<a href="%s" class="button wc-forward%s">%s</a> %s',
									wc_get_cart_url(),
									esc_attr( $wp_button_class ),
									__( 'View cart', 'woocommerce-product-batch-numbers' ),
									/* translators: 1: quantity in stock 2: current quantity */
									sprintf( __( 'You cannot add that amount to the cart &mdash; we have %1$s in stock and you already have %2$s in your cart.', 'woocommerce-product-batch-numbers' ), wpo_wcpbn_format_stock_amount( $qty_limit ), wpo_wcpbn_format_stock_amount( $stock_quantity_in_cart ) )
								);
							}
						}
						
					}
					
				}
				
				if ( $message && $page ) {
					switch ( $page ) {
						case 'product':
							throw new \Exception( $message );
							break;
						case 'cart':
							wc_add_notice( $message, 'error' );
							return true;
					}
				}
				
			}
			
			return false;
		}

	}

}