<?php

if ( ! defined( 'ABSPATH' ) ) {
	exit;
}

/**
 * Get batch numbers expiry date for a given order item.
 *
 * @return void|string
 */
function wpo_wcpbn_get_item_batch_expiry_dates( $item, $sep = ', ', $before = '', $after = '' ) {
	// check if we have an item
	if ( empty( $item ) ) {
		return;
	}
	
	// check if we have meta
	if ( empty( $item_batch_numbers = $item->get_meta( '_wcpbn_data' ) ) ) {
		return;
	}

	$expiry_dates = array();
	foreach ( $item_batch_numbers as $batch_number ) {
		$batch = wpo_wcpbn_get_batches( [ 'id' => $batch_number['id'] ] );
		if ( empty( $batch ) ) {
			continue;
		}
		$batch       = is_array( $batch ) ? reset( $batch ) : $batch;
		$date_expiry = '';
		if ( $batch->date_expiry != '0000-00-00 00:00:00' && $batch->date_expiry != '1000-01-01 00:00:00' ) {
			$date_expiry = get_date_from_gmt( $batch->date_expiry, wpo_wcpbn_date_format() );
		}
		$expiry_dates[] = $date_expiry;
	}

	return $before . join( $sep, $expiry_dates ) . $after;
}

/**
 * Get batch numbers for a given order item.
 *
 * @return void|string
 */
function wpo_wcpbn_get_item_batch_numbers( $item, $sep = ', ', $before = '', $after = '' ) {
	// check if we have an item
	if ( empty( $item ) ) {
		return;
	}

	// check if item has meta
	if ( empty( $item_batch_numbers = $item->get_meta( '_wcpbn_data' ) ) ) {
		return;
	}

	$batch_numbers = array();
	foreach ( $item_batch_numbers as $batch_number ) {
		$batch_numbers[] = $batch_number['batch_number'];
	}

	return $before . join( $sep, $batch_numbers ) . $after;
}

/**
 * Get batches.
 * 
 * @param  array $args
 *
 * @return null|array
 */
function wpo_wcpbn_get_batches( $args = [] ) {
	/*
	$args = [
		'id'                 => 1,
		'date_created'       => 'TIMESTAMP...TIMESTAMP', // more examples here: https://github.com/woocommerce/woocommerce/wiki/wc_get_orders-and-WC_Order_Query#date
		'date_expiry'        => 'TIMESTAMP...TIMESTAMP',
		'user_id'            => 1,
		'shared_ids'         => [ 23, 12 ],
		'batch_number'       => 'WOAIWPWPAP',
		'supplier'           => 'WP Overnight',
		'location'           => 'A83',
		'quantity'           => 12,
		'quantity_available' => 23,
		'status'             => [ 'active', 'archived' ], // can be a string as well: 'active'
		'search'             => [ 'keys' => [ 'batch_number', 'supplier' ], 'term' => 'WPO', 'type' => 'exact' ], // can only search on this keys: 'batch_number' and/or 'supplier'. Type can be 'exact' or by 'pattern'.
		'order_by'           => 'date_expiry',
		'order'              => 'ASC',
		'limit'              => 25,
		'offset'             => 5,
	];
	*/
	
	$join_callback = [];
	if ( ! empty( $args['shared_ids'] ) ) {
		if ( in_array( gettype( $args['shared_ids'] ), [ 'string', 'integer' ] ) ) {
			$args['shared_ids'] = [ intval( $args['shared_ids'] ) ];
		}
		$join_callback = [
			WPO_WCPBN()->classes['database']->shared_products(),
			'join_callback'
		];
	}
	
	return WPO_WCPBN()->classes['database']->batch_numbers()->get_rows( $args, $join_callback );
}

/**
 * Update batch.
 * 
 * @param  int|string|array $id
 * @param  array            $args
 *
 * @return bool
 */
function wpo_wcpbn_update_batch( $id, $args = [] ) {
	if ( empty( $id ) ) {
		return false;
	}
	
	if ( ! is_array( $id ) ) {
		$where = [ 'id' => intval( $id ) ];
	} else {
		$where = $id;
	}
	
	/*
	$args = [
		'date_expiry'        => '2023-01-16',
		'shared_ids'         => [ 23, 12 ],
		'batch_number'       => 'WOAIWPWPAP',
		'supplier'           => 'WP Overnight',
		'location'           => 'A83',
		'quantity_available' => 23,
		'units'              => 1,
		'status'             => 'active',
		'offset'             => 5,
	];
	*/
	
	if ( ! empty( $args['shared_ids'] ) && is_array( $args['shared_ids'] ) ) {
		$shared_ids = wpo_wcpbn_get_batch_shared_product_ids( $id );
		$diff       = wpo_wcpbn_batch_shared_products_diff( $shared_ids, $args['shared_ids'] );
		
		wpo_wcpbn_add_products_to_batch( $id, $diff['new'] );          // add new
		wpo_wcpbn_remove_products_from_batch( $id, $diff['deleted'] ); // remove deleted
		unset( $args['shared_ids'] );
	}

	return WPO_WCPBN()->classes['database']->batch_numbers()->update_row( $args, $where );
}

/**
 * @param array<int, array<string, mixed>> $args
 *
 * Sample input:
 * $args = array( 1 => array( 'date_expiry' => '2023-01-16', 'batch_number' => 'WOAIWPWPAP' ) );
 *
 * @return bool
 */
function wpo_wcpbn_update_batches( array $args ): bool {
	if ( empty( $args ) ) {
		return false;
	}

	// Unset shared_ids to prevent high loads of query execution.
	foreach ( $args as $key => $arg ) {
		unset( $args[ $key ]['shared_ids'] );
	}

	return WPO_WCPBN()->classes['database']->batch_numbers()->update_rows( $args );
}

/**
 * Insert batch.
 * 
 * @param  array $args
 *
 * @return bool
 */
function wpo_wcpbn_insert_batch( $args = [] ) {
	if ( empty( $args ) ) {
		return false;
	}
	
	/*
	$args = array(
		'date_created'       => '2023-01-16',
		'date_expiry'        => '2023-01-16',
		'shared_ids'         => [ 23, 12 ],
		'user_id'            => 1,
		'batch_number'       => 'WOAIWPWPAP',
		'supplier'           => 'WP Overnight',
		'location'           => 'A83',
		'quantity'           => 12,
		'quantity_available' => 23,
		'units'              => 1,
		'status'             => 'active',
		'offset'             => 5,
	);
	*/
	
	$shared_ids = [];
	if ( ! empty( $args['shared_ids'] ) ) {
		$shared_ids = $args['shared_ids'];
		unset( $args['shared_ids'] );
	}
	
	$inserted_id = WPO_WCPBN()->classes['database']->batch_numbers()->insert_row( $args );
	
	if ( $inserted_id ) {
		if ( ! empty( $shared_ids ) && is_array( $shared_ids ) ) {
			wpo_wcpbn_add_products_to_batch( $inserted_id, $shared_ids );
		}
		return $inserted_id;
	} else {
		return false;
	}
}

/**
 * Get batch supplier by name.
 * 
 * @param  string $name
 *
 * @return array|bool
 */
function wpo_wcpbn_get_batch_supplier_by_name( $name ) {
	if ( empty( $name ) ) {
		return;
	}

	return WPO_WCPBN()->classes['database']->batch_numbers()->get_supplier_by_name( $name );
}

/**
 * Get batch order IDs.
 * 
 * @param  string|int $id
 *
 * @return array|bool
 */
function wpo_wcpbn_get_batch_order_ids( $id ) {
	if ( empty( $id ) ) {
		return [];
	}

	// Get active batches
	$order_ids = WPO_WCPBN()->classes['database']->batch_numbers()->get_batch_order_ids( $id );

	if ( empty( $order_ids ) ) {
		return [];
	} else {
		$order_ids = reset( $order_ids );
	}

	$order_ids = json_decode( $order_ids->order_ids );

	return $order_ids;
}

/**
 * Restores or reduces the batch number available quantity.
 *
 * @return void
 */
function wpo_wcpbn_update_batch_stock_available_qty( $id, $value, $target = '' ) {
	if ( empty( $id ) || empty( $value ) || $value < 0 ) {
		return;
	}

	$value = wpo_wcpbn_force_stock_amount_type( $value );
	$batch = wpo_wcpbn_get_batches( [ 'id' => $id ] );
	if ( empty( $batch ) ) {
		return;
	}
	$batch = is_array( $batch ) ? reset( $batch ) : $batch;

	if ( empty( $target ) ) {
		$data['quantity_available'] = $value;
	} elseif ( $target == 'restore' ) {
		$data['quantity_available'] = $batch->quantity_available + $value;
	} elseif ( $target == 'reduce' ) {
		$reduce_qty    = $batch->quantity_available - $value;
		$qty_available = $reduce_qty < 0 ? 0 : $reduce_qty; // don't allow to go below zero
		$data['quantity_available'] = $qty_available;
	}

	// change status if quantity available reduces to zero
	if ( $data['quantity_available'] == 0 ) {
		$data['status'] = 'archived';
	}

	wpo_wcpbn_update_batch( $id, $data );
}

/**
 * Delete batch.
 *
 * @return void
 */
function wpo_wcpbn_delete_batch( $id, $data = [], $force = false ) {
	if ( empty( $id ) ) {
		return;
	}
	
	$messages = [];
	
	// maybe log messages
	if ( wc_string_to_bool( WPO_WCPBN()->classes['settings']->options->get_option( 'log_batch_deletion' ) ) ) {
		$logger          = \WPO\WC\Product_Batch_Numbers\wcpbnLogger::instance();
		$logger->context = [ 'source' => 'wpo-wcpbn-deletions' ];
	} else {
		$logger = false;
	}
	
	if ( ! empty( $data ) ) {
		extract( $data );
	}

	$batch = wpo_wcpbn_get_batches( [ 'id' => intval( $id ) ] );
	if ( empty( $batch ) ) {
		$messages = [
			/* translators: batch ID */
			'error' => sprintf( __( 'Batch with ID %s not found!', 'woocommerce-product-batch-numbers' ), intval( $id ) ),
		];
		if ( $logger ) {
			$logger->log( $messages['error'], 'error' );
		}
		WPO_WCPBN()->classes['admin_notices']->save( $messages['error'], 'error' );
		return;
	}
	$batch          = is_array( $batch ) ? reset( $batch ) : $batch;
	$all_shared_ids = wpo_wcpbn_get_batch_shared_product_ids( $batch->id );
	$messages       = [
		'error' => __( 'An error occurred while trying to delete a batch.', 'woocommerce-product-batch-numbers' ),
	];
	$can_delete     = $force;
	$deleted        = false;
	
	// safe delete validation
	if ( count( $all_shared_ids ) > 0 ) {
		if ( empty( $product_id ) ) {
			$messages['error'] = sprintf(
				/* translators: 1. batch number, 2. batch ID */
				__( 'The batch %1$s (ID: #%2$s) cannot be deleted because it is being shared with at least one product.', 'woocommerce-product-batch-numbers' ),
				$batch->batch_number,
				$id
			);
		} else {
			// remove product
			$product_removed = wpo_wcpbn_remove_products_from_batch( $batch->id, [ $product_id ] );
			// go again
			wpo_wcpbn_delete_batch( $id );
			return;
		}
	} else {
		$can_delete = true;
	}
	
	// batch deletion
	if ( apply_filters( 'wpo_wcpbn_force_batch_deletion', $can_delete, $id ) ) {
		$deleted  = WPO_WCPBN()->classes['database']->batch_numbers()->delete_row( [ 'id' => $id ] );
		$messages = [
			'success' => sprintf(
				/* translators: 1. batch number, 2. batch ID */
				__( 'The batch %1$s (ID: #%2$s) was deleted.', 'woocommerce-product-batch-numbers' ),
				$batch->batch_number,
				$id
			),
			'error' => sprintf( 
				/* translators: 1. batch number, 2. batch ID */
				__( 'An error occurred while removing the batch %1$s (ID: #%2$s).', 'woocommerce-product-batch-numbers' ),
				$batch->batch_number,
				$id
			),
		];
	}
	
	if ( $deleted ) {
		if ( $logger ) {
			$logger->log( $messages['success'], 'notice' );
		}
		if ( apply_filters( 'wpo_wcpbn_show_batch_deletion_notices', false ) ) {
			WPO_WCPBN()->classes['admin_notices']->save( $messages['success'], 'success' );
		}
	} else {
		if ( $logger ) {
			$logger->log( $messages['error'], 'error' );
		}
		WPO_WCPBN()->classes['admin_notices']->save( $messages['error'], 'error' );
	}
}

/**
 * Remove products from batch
 *
 * @return bool
 */
function wpo_wcpbn_remove_products_from_batch( $id, $product_ids = [] ) {
	if ( empty( $id ) || empty( $product_ids ) ) {
		return false;
	}
	
	$error = 0;
	
	foreach ( $product_ids as $product_id ) {
		$args = [
			'batch_id'   => intval( $id ),
			'product_id' => intval( $product_id ),
		];
		$row = WPO_WCPBN()->classes['database']->shared_products()->get_rows( $args );
		$row = ! empty( $row ) ? reset( $row ) : false;
		
		if ( $row ) {
			$query = WPO_WCPBN()->classes['database']->shared_products()->delete_row( [ 'id' => $row->id ] );
			
			// sync WC stock
			$product = wc_get_product( intval( $product_id ) );
			if ( $product ) {
				wpo_wcpbn_sync_product_stock_with_batches( $product );
			}
			
			if ( ! $query ) {
				$error++;
				WPO_WCPBN()->classes['logger']->log(
					sprintf( 'An error occurred while deleting the product #%1$s from the batch #%2$s.' ),
					esc_attr( $product_id ),
					esc_attr( $id )
				);
			}
		}
	}
	
	if ( $error > 0 ) {
		return false;
	} else {
		return true;
	}
}

/**
 * Add products to batch
 * 
 * @param  int|string $id           batch ID
 * @param  array      $product_ids
 *
 * @return bool
 */
function wpo_wcpbn_add_products_to_batch( $id, $product_ids = [] ) {
	if ( empty( $id ) || empty( $product_ids ) ) {
		return false;
	}
	
	$error = 0;
	
	foreach ( $product_ids as $product_id ) {
		$args = [
			'date_created' => date( 'Y-m-d H:i:s' ),
			'batch_id'     => intval( $id ),
			'product_id'   => intval( $product_id ),
		];
		$query = WPO_WCPBN()->classes['database']->shared_products()->insert_row( $args );
		
		// sync WC stock
		$product = wc_get_product( intval( $product_id ) );
		if ( $product ) {
			wpo_wcpbn_sync_product_stock_with_batches( $product );
		}
		
		if ( ! $query ) {
			$error++;
			WPO_WCPBN()->classes['logger']->log(
				sprintf( 'An error occurred while adding the product #%1$s to the batch #%2$s.' ),
				esc_attr( $product_id ),
				esc_attr( $id )
			);
		}
	}
	
	if ( $error > 0 ) {
		return false;
	} else {
		return true;
	}
}

/**
 * Checks if a product has batches
 *
 * @return bool
 */
function wpo_wcpbn_product_has_batches( $product_id ) {
	$has_batches = false;

	if ( empty( $product_id ) ) {
		return $has_batches;
	}
	
	$query = WPO_WCPBN()->classes['database']->shared_products()->get_rows( [ 'product_id' => intval( $product_id ) ] );

	if ( $query && count( $query ) > 0 ) {
		$has_batches = true;
	}

	return $has_batches;
}

/**
 * Get product batches stock total.
 *
 * @return int
 */
function wpo_wcpbn_get_product_batches_stock_total( $product_id ) {
	$stock_total = 0;

	if ( empty( $product_id ) ) {
		return $stock_total;
	}
	
	$args = [
		'shared_ids' => [ intval( $product_id ) ],
		'status'     => 'active',
	];
	
	$batches = wpo_wcpbn_get_batches( $args );

	if ( empty( $batches ) ) {
		return $stock_total;
	}

	foreach ( $batches as $batch ) {
		$stock_total += wpo_wcpbn_force_stock_amount_type( $batch->quantity_available );
	}

	return $stock_total;
}

/**
 * Syncs the product stock with the batches stock
 *
 * @return bool
 */
function wpo_wcpbn_sync_product_stock_with_batches( $product ) {
	$synced = false;

	if ( empty( $product ) || ! is_object( $product ) ) {
		return $synced;
	}

	if ( ! $product->managing_stock() ) {
		return $synced;
	}

	if ( ! wpo_wcpbn_product_has_batches( $product->get_id() ) ) {
		return $synced;
	}
	
	// mirror enabled?
	$is_mirroring = wpo_wcpbn_is_product_mirroring_batches_stock( $product->get_id() );

	// use batches total stock
	if ( ! $is_mirroring ) {
		$stock_to_update = wpo_wcpbn_get_product_batches_stock_total( $product->get_id() );
	// use the batches highest stock value
	} else {
		$stock_to_update = wpo_wcpbn_get_product_batches_highest_stock_value( $product->get_id() );
		wpo_wcpbn_sync_product_batches_stock_available( $product->get_id(), $stock_to_update );
	}

	if ( $stock_to_update >= 0 ) {
		// if WC stock sync setting is enabled
		if ( wc_string_to_bool( WPO_WCPBN()->classes['settings']->options->get_option( 'sync_wc_stock' ) ) ) {
			// update stock
			wc_update_product_stock( $product, wpo_wcpbn_force_stock_amount_type( $stock_to_update ), 'set', false );
			$synced = true;
		}

		// update stock status
		wpo_wcpbn_set_product_stock_status( $product->get_id() );
	}

	return $synced;
}

/**
 * Syncs the product batches stock availability for a given value
 *
 * @return bool
 */
function wpo_wcpbn_sync_product_batches_stock_available( $product_id, $stock_value ) {
	if ( empty( $product_id ) || empty( $stock_value ) || ! is_numeric( $stock_value ) ) {
		return false;
	}

	$args = [
		'shared_ids' => [ intval( $product_id ) ],
		'status'     => 'active',
	];
	
	$batches = wpo_wcpbn_get_batches( $args );
	
	$error = 0;

	if ( ! empty( $batches ) ) {
		foreach ( $batches  as $batch ) {
			if ( is_array( $batch ) ) {
				$batch = (object) $batch;
			}
			
			$query = wpo_wcpbn_update_batch_stock_available_qty( $batch->id, wpo_wcpbn_force_stock_amount_type( $stock_value ) );
			
			if ( ! $query ) {
				$error++;
			}
		}
		
	}
	
	if ( $error > 0 ) {
		return false;
	} else {
		return true;
	}
}

/**
 * Gets the highest stock value from product batches
 *
 * @return int
 */
function wpo_wcpbn_get_product_batches_highest_stock_value( $product_id ) {
	$highest = 0;

	if ( empty( $product_id ) ) {
		return $highest;
	}

	$args = [
		'shared_ids' => [ intval( $product_id ) ],
		'status'     => 'active',
	];
	
	$batches = wpo_wcpbn_get_batches( $args );

	if ( ! empty( $batches ) ) {
		$highest = wpo_wcpbn_force_stock_amount_type( max( array_column( $batches, 'quantity_available' ) ) );
	}

	return $highest;
}

/**
 * Check if a product is mirroring batches stock on setting
 *
 * @return bool
 */
function wpo_wcpbn_is_product_mirroring_batches_stock( $product_id ) {
	$is_mirroring = false;

	if ( empty( $product_id ) ) {
		return $is_mirroring;
	}

	$product = wc_get_product( $product_id );

	if ( ! empty( $product ) && $product->managing_stock() ) {
		$is_mirroring = wc_string_to_bool( $product->get_meta( '_wcpbn_mirror_batches_stock' ) );
	}

	return $is_mirroring;
}

/**
 * Gets the global offset in days
 *
 * @return int
 */
function wpo_wcpbn_get_global_offset_in_days() {
	$offset = WPO_WCPBN()->classes['settings']->options->get_option( 'global_expiration_offset' );
	
	if ( empty( $offset ) ) {
		$offset = apply_filters( 'wpo_wcpbn_default_offset_value', '30' );
	}
	
	return $offset;
}

/**
 * Gets the product current offset in days
 *
 * @return int
 */
function wpo_wcpbn_get_product_current_offset_in_days( $product_id ) {
	if ( empty( $product_id ) ) {
		return 0;
	}
	
	// global offset
	$offset = wpo_wcpbn_get_global_offset_in_days();

	$product = wc_get_product( $product_id );
	if ( ! empty( $product ) ) {
		$product_offset = $product->get_meta( '_wcpbn_notification_offset', true );
		if ( ! empty( $product_offset ) && is_numeric( $product_offset ) ) {
			$offset = intval( $product_offset );
		}
	}
	
	return $offset;
}

/**
 * Get product last added batch.
 *
 * @return bool|object
 */
function wpo_wcpbn_get_product_last_added_batch( $product_id, $include_archived = false ) {
	if ( empty( $product_id ) ) {
		return false;
	}

	$args = [
		'shared_ids' => [ absint( $product_id ) ],
		'status'     => ( $include_archived ) ? [ 'active', 'archived' ] : 'active',
	];
	
	$batches = wpo_wcpbn_get_batches( $args );

	if ( ! empty( $batches ) && is_array( $batches ) ) {
		rsort( $batches );
		return reset( $batches );
	}

	return false;
}

/**
 * Get product batch data formatted.
 *
 * @return string
 */
function wpo_wcpbn_get_product_batch_data_formatted( $product_id, $data_format, $include_quantity = false, $include_label = true, $days_from = '', $days_to = '', $limit = '', $batches = '' ) {
	if ( empty( $product_id ) || empty( $data_format ) || $data_format == 'no' ) {
		return;
	}

	$product = wc_get_product( $product_id );
	if ( empty( $product ) ) {
		return;
	}

	if ( empty( $batches ) ) {
		$args = array(
			'shared_ids' => [ $product->get_id() ],
			'status'     => 'active',
			'order_by'   => WPO_WCPBN()->classes['settings']->options->get_option( 'batch_prioritization' ) ? WPO_WCPBN()->classes['settings']->options->get_option( 'batch_prioritization' ) : 'id',
			'order'      => 'ASC',
		);
		
		$batches = wpo_wcpbn_get_batches( $args );
		
		if ( empty( $batches ) ) {
			return;
		}
	}

	if ( $limit && is_numeric( $limit ) ) {
		$batches = array_slice( $batches, 0, intval( $limit ) );
	}

	$label = apply_filters( 'wpo_wcpbn_product_batches_data_formatted_label', __( 'Batches', 'woocommerce-product-batch-numbers' ), $product );
	
	$formatted = '<div class="wpo-wcpbn-product-batch-data">';
	if ( $include_label ) {
		$formatted .= '<label>'.$label.':</label>';
	}
	$formatted .= '<ul>';
	foreach ( $batches as $batch ) {
		if ( is_array( $batch ) ) {
			$batch = (object) $batch;
		}

		$include_li = false;
		$li         = '<li>';

		// prepare date
		if ( isset( $batch->date_expiry ) && $batch->date_expiry != '0000-00-00 00:00:00' && $batch->date_expiry != '1000-01-01 00:00:00' ) {
			$today      = time();
			$batch_date = strtotime( $batch->date_expiry );
			if ( ! empty( $days_from ) && is_numeric( $days_from ) ) {
				$show_date  = strtotime( '-' . intval( $days_from ) . ' days', $batch_date );
			}
			if ( ! empty( $days_to ) && is_numeric( $days_to ) ) {
				$hide_date  = strtotime( '-' . intval( $days_to ) . ' days', $batch_date );
			}
			
			$class = wpo_wcpbn_get_batch_expiry_date_class( $batch );
			if ( $include_label ) {
				$date_label = __( 'Expires on', 'woocommerce-product-batch-numbers' ) . ' ' . date_i18n( wpo_wcpbn_date_format(), $batch_date );
			} else {
				$date_label = date_i18n( wpo_wcpbn_date_format(), $batch_date );
			}
			$date_label = apply_filters( 'wpo_wcpbn_product_batches_data_formatted_date_label', $date_label, $batch );

			// conditions
			if ( isset( $show_date ) && isset( $hide_date ) ) {						
				if ( $today <= $batch_date && $show_date <= $today && $today <= $hide_date  ) {
					$include_li = true;
				}
			} elseif ( ! isset( $show_date ) && isset( $hide_date ) ) {
				if ( $today <= $batch_date && $today <= $hide_date ) {
					$include_li = true;
				}
			} elseif ( isset( $show_date ) && ! isset( $hide_date ) ) {
				if ( $today <= $batch_date && $show_date <= $today ) {
					$include_li = true;
				}
			} elseif ( ! isset( $show_date ) && ! isset( $hide_date ) ) {
				if ( $today <= $batch_date ) {
					$include_li = true;
				}
			}
		}

		switch ( $data_format ) {
			case 'number':
				if ( ! empty( $batch->batch_number ) ) {
					$li .= '<span class="number">'.$batch->batch_number.'</span>';
				}
				if ( $include_quantity ) {
					if ( isset( $batch->quantity_sold ) ) {
						$li .= '<span class="quantity-sold">'.wpo_wcpbn_format_stock_amount( $batch->quantity_sold ).'x</span>';
					} else {
						$li .= '<span class="quantity-available">'.wpo_wcpbn_format_stock_amount( $batch->quantity_available ).'x</span>';
					}
				}
				$include_li = true;
				break;
			case 'date':
				if ( $include_quantity ) {
					if ( isset( $batch->quantity_sold ) ) {
						$li .= '<span class="quantity-sold">'.wpo_wcpbn_format_stock_amount( $batch->quantity_sold ).'x</span>';
					} else {
						$li .= '<span class="quantity-available">'.wpo_wcpbn_format_stock_amount( $batch->quantity_available ).'x</span>';
					}
				}
				if ( isset( $class ) && isset( $date_label ) ) {
					$li .= sprintf( '<span class="expiry-date %1$s">%2$s</span>', $class, $date_label );
				}
				$include_li = true;
				break;
			case 'number+date':
				if ( ! empty( $batch->batch_number ) ) {
					$li .= '<span class="number">'.$batch->batch_number.'</span>';
				}
				if ( $include_quantity ) {
					if ( isset( $batch->quantity_sold ) ) {
						$li .= '<span class="quantity-sold">'.wpo_wcpbn_format_stock_amount( $batch->quantity_sold ).'x</span>';
					} else {
						$li .= '<span class="quantity-available">'.wpo_wcpbn_format_stock_amount( $batch->quantity_available ).'x</span>';
					}
				}
				if ( isset( $class ) && isset( $date_label ) ) {
					$li .= sprintf( '<span class="expiry-date %1$s">%2$s</span>', $class, $date_label );
				}
				$include_li = true;
				break;
		}

		$li .= '</li>';

		if ( $include_li ) {
			$formatted .= $li;
		}
	}
	$formatted .= '</ul></div>';
	
	return apply_filters( 'wpo_wcpbn_product_batches_data_formatted', $formatted, $product );
}

/**
 * Print product batch data formatted.
 *
 * @return void
 */
function wpo_wcpbn_print_product_batch_data_formatted( $product_id, $data_format, $include_quantity = false, $include_label = true, $days_from = '', $days_to = '', $limit = '', $batches = '' ) {
	echo wpo_wcpbn_get_product_batch_data_formatted( $product_id, $data_format, $include_quantity, $include_label, $days_from, $days_to, $limit, $batches );
}

/**
 * Get all batch numbers matching givin string.
 *
 * @return void|array
 */
function wpo_wcpbn_get_all_batch_numbers_matching_string( $string ) {
	if ( empty( $string ) ) {
		return;
	}

	// Get all batches matching
	$batches = wpo_wcpbn_get_batches( [ 'batch_number' => $string ] );

	if ( empty( $batches ) ) {
		return;
	}

	return $batches;
}

/**
 * Get batches between date range.
 *
 * @return array
 */
function wpo_wcpbn_get_batches_between_date_range( $date_from, $date_to, $status = 'all' ) {
	$args = array();

	if ( ! empty( $status ) ) {
		if ( $status == 'all' ) {
			$args['status'] = array( 'active', 'archived' );
		} else {
			$args['status'] = $status;
		}
	}

	if ( ! empty( $date_from ) && ! empty( $date_to ) ) {
		$args['date_created'] = "{$date_from}...{$date_to}";
	} elseif ( ! empty( $date_from ) ) {
		$today = time();
		$args['date_created'] = "{$date_from}...{$today}";
	} elseif ( ! empty( $date_to ) ) {
		$args['date_created'] = "<={$date_to}";
	}

	$batches = wpo_wcpbn_get_batches( $args );

	if ( empty( $batches ) ) {
		return array();
	}

	return $batches;
}

/**
 * Set product stock status.
 * 
 * @param $id
 *
 * @return void
 */
function wpo_wcpbn_set_product_stock_status( $id ) {
	if ( ! empty( $id ) ) {
		$product = wc_get_product( $id );
		if ( ! empty( $product ) ) {
			if ( $product->managing_stock() ) {
				$current_stock = $product->get_stock_quantity();
			} else {
				$current_stock = wpo_wcpbn_get_product_batches_stock_total( $product->get_id() );
			}

			$current_stock_status = $product->get_stock_status();

			// backorder disabled
			if ( ! $product->is_on_backorder() ) {
				if ( $current_stock > 0 ) {
					$product->set_stock_status( 'instock' );
				} else {
					$product->set_stock_status( 'outofstock' );
				}
			// backorder enabled
			} else {
				if ( $current_stock > 0 && $current_stock_status != 'instock' ) {
					$product->set_stock_status( 'instock' );
				} else {
					$product->set_stock_status( 'onbackorder' );
				}
			}

			// save data
			$product->save();
		}
	}
}

/**
 * Allocate batch numbers for a product/item quantity.
 *
 * @return array
 */
function wpo_wcpbn_allocate_batch_numbers( $product_id, $quantity, $item = null, $order_id = null ) {
	return WPO_WCPBN()->classes['order_items']->allocate_batch_numbers( $product_id, $quantity, $item, $order_id );
}

/**
 * Gets the batch expiry date class from offset settings.
 *
 * @return string
 */
function wpo_wcpbn_get_batch_expiry_date_class( $batch ) {
	$class = '';
	
	if ( empty( $batch ) ) {
		return $class;
	}
	
	$batch      = is_array( $batch ) ? reset( $batch ) : $batch;
	$today      = time();
	$batch_date = strtotime( $batch->date_expiry );
	
	// global offset
	$offset = wpo_wcpbn_get_global_offset_in_days();
	
	if ( ! empty( $offset ) && ! empty( $batch_date ) ) {
		$less_x_days = strtotime( '-'.$offset.' days', $batch_date );

		if ( $batch_date < $today ) {
			$class = 'batch-expired';
		} elseif ( $less_x_days < $today ) {
			$class = 'batch-within-expiry-offset';
		} else {
			$class = 'batch-valid';
		}
	}

	return $class;
}

/**
 * Gets the batch shared product IDs.
 * 
 * @param int|string $batch_id
 *
 * @return array
 */
function wpo_wcpbn_get_batch_shared_product_ids( $id ) {
	if ( ! empty( $id ) ) {
		$query = WPO_WCPBN()->classes['database']->shared_products()->get_rows( [ 'batch_id' => $id ] );
		if ( $query ) {
			return array_map( function( $o ) { return intval( $o->product_id ); }, $query );
		} else {
			return [];
		}
	} else {
		return [];
	}
}

/**
 * Gets the shared product info by ID.
 * 
 * @param int|string $id
 * @param int|string $product_id
 *
 * @return array
 */
function wpo_wcpbn_get_batch_shared_product( $id, $product_id ) {
	if ( empty( $id ) || empty( $product_id ) ) {
		return [];
	}
	
	$args = [
		'batch_id'   => intval( $id ),
		'product_id' => intval( $product_id ),
	];
	
	return WPO_WCPBN()->classes['database']->shared_products()->get_rows( $args );
}

/**
 * Compared 2 arrays of product IDs.
 * 
 * @param array   $old_product_ids
 * @param array   $new_product_ids
 * @param string  $output           can be: all, new, unchanged, deleted
 * 
 * @return array
 */
function wpo_wcpbn_batch_shared_products_diff( $old_product_ids, $new_product_ids, $output = 'all' ) {
	$new    = array_flip( $new_product_ids );
	$old    = array_flip( $old_product_ids );
	$update = array_intersect_key( $old, $new );
	$delete = array_diff_key( $old, $update );
	$create = array_diff_key( $new, $update );
	
	$diff = [
		'new'       => array_keys( $create ),
		'unchanged' => array_keys( $update ),
		'deleted'   => array_keys( $delete )
	];
	
	switch ( $output ) {
		default:
		case 'all':
			return $diff;
		case 'new':
			return $diff['new'];
		case 'unchanged':
			return $diff['unchanged'];
		case 'deleted':
			return $diff['deleted'];
	}
}

/**
 * WPML object ID compatibility.
 *
 * @return int
 */
function wpo_wcpbn_wpml_object_id( $id ) {
	if ( empty( $id ) ) {
		return 0;
	}
	
	$wpml_language = apply_filters( 'wpo_wcpbn_wpml_language', apply_filters( 'wpml_default_language', null ), $id );

	return apply_filters( 'wpml_object_id', $id, get_post_type( $id ), true, $wpml_language );
}

/**
 * Gets the stock amount type from WC filter
 * 
 * WC stock amounts are integers by default.
 *
 * @return int|float
 */
function wpo_wcpbn_force_stock_amount_type( $amount ) {
	if ( ! is_numeric( $amount ) ) {
		return $amount;
	}

	if ( ! has_filter( 'woocommerce_stock_amount', 'intval' ) ) {
		return floatval( wc_format_decimal( $amount ) );
	} else {
		return intval( $amount );
	}
}

/**
 * Formats the stock amount to be displayed
 *
 * @return string
 */
function wpo_wcpbn_format_stock_amount( $amount ) {
	if ( ! is_numeric( $amount ) ) {
		return $amount;
	}

	$amount = wpo_wcpbn_force_stock_amount_type( $amount );

	if ( ! has_filter( 'woocommerce_stock_amount', 'intval' ) ) {
		return wc_format_localized_decimal( strval( $amount ) );
	} else {
		return strval( $amount );
	}
}

/**
 * Returns de date format
 *
 * @return string
 */
function wpo_wcpbn_date_format() {
	$format        = wc_date_format();
	$format        = apply_filters_deprecated( 'wpo_wcpbn_format_date', array( $format ), '1.3.0', 'wpo_wcpbn_date_format' );
	$custom_format = WPO_WCPBN()->classes['settings']->options->get_option( 'custom_date_format' );

	// custom date format
	if ( ! empty( $custom_format ) && $custom_format != $format ) {
		$format = $custom_format;
	}

	return apply_filters( 'wpo_wcpbn_date_format', $format );
}

/**
 * Get batch ids for product ids
 * 
 * @param $product_ids array of product ids
 * 
 * @return array
 */ 
function wpo_wcpbn_get_product_batch_ids( $product_ids ) {
	if ( ! empty( $product_ids ) ) {
		$query = WPO_WCPBN()->classes['database']->shared_products()->get_rows( [ 'product_id' => $product_ids ] );
		if ( $query ) {
			return array_map( function( $o ) { return intval( $o->batch_id ); }, $query );
		} else {
			return [];
		}
	} else {
		return [];
	}
}

/**
 * Get purchase order for batch
 * 
 * @param int $batch_id
 * 
 * @return array|object
 */ 
function wpo_wcpbn_get_batch_purchase_order( int $batch_id ) {
	$po = array();
	
	if ( ! empty( $batch_id ) && 0 < $batch_id ) {
		$query = WPO_WCPBN()->classes['database']->purchase_orders()->get_row_by_batch_id( absint( $batch_id ) );
		if ( $query ) {
			$po = reset( $query );
		}
	}
	
	return $po;
}

/**
 * Insert batch purchase order.
 * 
 * @param  array $args
 *
 * @return bool
 */
function wpo_wcpbn_insert_batch_purchase_order( $args ) {
	if ( empty( $args ) ) {
		return false;
	}
	
	/*
	$args = array(
		'date_created'   => '2023-01-16 00:00:00',
		'batch_id'       => '23',
		'po_number'      => 'PO2345',
		'supplier'       => 'WPO'
		'cost'           => 10.99,
		'regular_price'  => 14.99,
		'sale_price'     => 13.99,
		'price_margin'   => 30.00, // percentage
		'price_discount' => 5.00,  // percentage
	);
	*/
	
	$args = WPO_WCPBN()->classes['purchase_orders']->calculate_prices( $args );
	
	return WPO_WCPBN()->classes['database']->purchase_orders()->insert_row( $args );
}

/**
 * Update batch purchase order.
 * 
 * @param  int    $id
 * @param  array  $args
 *
 * @return bool
 */
function wpo_wcpbn_update_batch_purchase_order( int $id, array $args = array() ): bool {
	if ( 0 < $id ) {
		$where = array( 'id' => absint( $id ) );
	} else {
		return false;
	}
	
	/*
	$args = array(
		'date_created'   => '2023-01-16 00:00:00',
		'batch_id'       => '23',
		'po_number'      => 'PO2345',
		'supplier'       => 'WPO'
		'cost'           => 10.99,
		'regular_price'  => 14.99,
		'sale_price'     => 13.99,
		'price_margin'   => 30.00, // percentage
		'price_discount' => 5.00,  // percentage
	);
	*/

	$args = WPO_WCPBN()->classes['purchase_orders']->calculate_prices( $args );
	
	return WPO_WCPBN()->classes['database']->purchase_orders()->update_row( $args, $where );
}

/**
 * Delete batch purchase order.
 * 
 * @param  int $id
 *
 * @return bool
 */
function wpo_wcpbn_delete_batch_purchase_order( int $id ): bool {
	if ( 0 < $id ) {
		$where = array( 'id' => absint( $id ) );
	} else {
		return false;
	}

	return WPO_WCPBN()->classes['database']->purchase_orders()->delete_row( $where );
}

/**
 * Get purchase orders from batch IDs.
 * 
 * @param  array $batch_ids
 *
 * @return array
 */
function wpo_wcpbn_get_pos_from_batch_ids( array $batch_ids ): array {
	if ( empty( $batch_ids ) ) {
		return array();
	}
	
	$ids = array_map( 'intval', $batch_ids );
	
	return WPO_WCPBN()->classes['database']->purchase_orders()->get_rows_from_batch_ids( $ids );
}

